----------------------------------------------------------------------
----                                                              ----
---- ATARI Falcon COMBEL compatible IP Core	    	              ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- Atari's COMBEL with all features to reach                    ----
---- ATARI Falcon compatibility.                                  ----
----                                                              ----
---- Top level file for use in systems on programmable chips.     ----
----                                                              ----
---- To Do:                                                       ----
---- -                                                            ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright � 2009-2010 Wolfgang Foerster Inventronik GmbH.    ----
---- All rights reserved. No portion of this sourcecode may be    ----
---- reproduced or transmitted in any form by any means, whether  ----
---- by electronic, mechanical, photocopying, recording or        ----
---- otherwise, without my written permission.                    ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K9B  2009/12/24 WF
--   Initial Release.
-- 

library work;
use work.COMBEL_PKG.all;
library ieee;
use ieee.std_logic_1164.all;

entity COMBEL_TOP is
	port (
        -- System and core control:
        RESET 	            : in bit;
        CPU_CLKSEL_16_8n    : out bit;
        BLT_CLKSEL_16_8n    : out bit;
        CLK_32		        : in bit; -- Originally 32MHz.
        CLK_BLT             : in bit;
        CLK_8               : out bit;
        CLK_4               : out bit;
        KHz_500             : out bit; -- Clock for the MIDI ACIA and the keyboard ACIA (receiver).
        KHz_500W            : out bit; -- Clock for the transmitter of the keyboard ACIA.

        -- Adress and data bus:
        ADR_IN		        : in bit_vector(31 downto 1);
        ADR_16MB_UP         : in bit; -- Select the fully 32 bit address space.
        ADR_OUT		        : out std_logic_vector(31 downto 1);
        ADR_EN              : out bit; -- Used for the Blitter.
        DATA_IN		        : in std_logic_vector(15 downto 0);
        DATA_OUT	        : out std_logic_vector(15 downto 0);
        DATA_EN		        : out bit;

		-- The RAM interface:
        BANK0_TYPE          : out BANKTYPE;
        RAS1n	            : out bit;
        CAS1Hn	            : out bit;
        CAS1Ln	            : out bit;
        RAS0n	            : out bit;
        CAS0Hn	            : out bit;
        CAS0Ln	            : out bit;
        WEn	                : out bit;
        RAM_ADR 	        : out bit_vector(26 downto 1);

        -- Bus control:
        -- The signal BUS_EN enables the tri-state
        -- drivers of the following signals:
        -- RW_OUTn, AS_OUTn, UDS_OUTn, LDS_OUTn, 
        -- FC_OUT.
        BUS_EN	            : out bit;
		RWn_IN		        : in bit;
		RWn_OUT		        : out bit;
		AS_INn		        : in bit;
		AS_OUTn		        : out bit;
		UDS_INn		        : in bit;
		UDS_OUTn	        : out bit;
		LDS_INn		        : in bit;
		LDS_OUTn	        : out bit;

        DTACK_INn	        : in bit;
        DTACK_OUTn	        : out bit;

        -- 6800 peripheral control:
        VPAn			    : out bit;
        VMAn 		        : in bit;

        -- Bus status:
        BERRn		        : out bit;

        -- Processor function codes:
        FC_IN		        : in bit_vector(2 downto 0);
        FC_OUT		        : out std_logic_vector(2 downto 0);

        -- Bus arbitration control:
        BRn			        : out bit;
        BGIn			    : in bit;
        BGOn			    : out bit;
        BGACK_INn	        : in bit;
        BGACK_OUTn	        : out bit;

        -- Adress decoder stuff:
        -- In original COMBEL ther are only
        -- Pins for ROM2n, ROM3n and ROM4n.
        ROMSEL_FC_E0n       : in bit; -- Operating system's address space.
        ROM_6n	            : out bit;	-- STE.
        ROM_5n	            : out bit;	-- STE.
        ROM_4n	            : out bit;	-- ST.
        ROM_3n	            : out bit;	-- ST.
        ROM_2n	            : out bit;
        ROM_1n	            : out bit;
        ROM_0n	            : out bit;

        N6850	            : out bit;
        MFPCSn	            : out bit;

        SNDCS	            : out bit; -- STE: Sound chip select.
        SNDIR	            : out bit; -- STE: Data flow direction control.
        FPUCS               : out bit; -- Floating point unit
        R8006               : out bit; -- Falcon's configuration register.

        -- Keyboard stuff:
        TOK                 : in bit; -- 'Transmit Ok', to KROK of the Keyboard connector.
        TID                 : in bit; -- To TXD of the keyboard ACIA, not used yet.

        -- VIDEL control signals:
        VREQ                : in bit; -- Video data request.
        EVENn_ODD           : in bit; -- Indicates the interlaced video frame.
        VCS                 : out bit; -- VIDEL chip select.
        VLD                 : out bit; -- VIDEL's data latch control.
        RDAT                : out bit; -- VIDEL's data latch control.
        WDAT                : out bit; -- VIDEL's data latch control.
        RAMH			    : out bit; -- Indicates hi RAM access to the VIDEL.

        -- DMA control signal:
        BMODE               : in bit; -- '0' = 68030 bus timing '1' = 68000 bus timing.

        -- Real time clock:
        RTCCSn	            : out bit; -- Real time clock.
        RTC_WRn	            : out bit; -- Originally RTCASn (address strobe).
        RTC_RDn	            : out bit; -- Originally RTCDSn (data strobe).

        -- Interrupt system:
        HINTn		        : in bit; -- Horizontal interrupt.
        VINTn		        : in bit; -- Vertical Interrupt.
        MFPINTn		        : in bit;
        EINT1n	            : in bit;
        EINT3n	            : in bit;
        EINT5n	            : in bit;
        EINT7n	            : in bit;
        BINTn               : out bit; -- Blitter.
        AVECn		        : out bit; -- Add-On over COMBEL.
        IACKn		        : out bit; -- ST.
        IPL2n	            : out bit; -- STE only.
        IPL1n	            : out bit; -- STE only.
        IPL0n	            : out bit; -- STE only.

        -- IDE interface:
        IDE_RS0n            : out bit;
        IDE_RS1n            : out bit;
        IDE_IORDn           : out bit;
        IDE_IOWRn           : out bit;
        IDE_D_EN_INn        : out bit; -- In-Buffer control, Add-On over COMBEL.
        IDE_D_EN_OUTn       : out bit; -- Out-Buffer control, Add-On over COMBEL.

        -- SCC chip:
        SCCn		        : out bit;	-- Select signal for the STE or TT SCC chip.
        SCCAB               : out bit;
        SCCRD               : out bit;
        SCCWR               : out bit;
        SCCIACKn            : out bit;
        SCCWAITn            : in bit;

        -- Joyport:
        JOY_RHn	            : out bit; 		-- Read only FF9202 high byte.
        JOY_RLn	            : out bit; 		-- Read only FF9202 low byte.
        JOY_WL	            : out bit; 		-- Write only FF9202 low byte.
        JOY_WEn	            : out bit; 		-- Write only FF9202 output enable.
        BUTTONn	            : out bit; 		-- Read only FF9000 low byte.		
        PAD0Xn	            : in bit; 		-- Counter input for the Paddle 0X.
        PAD0Yn	            : in bit; 		-- Counter input for the Paddle 0Y.
        PAD1Xn	            : in bit; 		-- Counter input for the Paddle 1X.
        PAD1Yn	            : in bit; 		-- Counter input for the Paddle 1Y.
        PADRSTn	            : out bit 		-- Paddle monoflops reset.
    );
end entity COMBEL_TOP;
	
architecture STRUCTURE of COMBEL_TOP is
signal BERR_In                  : bit;
signal ADR_IN_I                 : bit_vector(31 downto 1);
signal DATA_OUT_BLITTER         : std_logic_vector(15 downto 0);
signal DATA_EN_BLITTER          : bit;
signal DATA_OUT_MCU		        : std_logic_vector(15 downto 0);
signal DATA_EN_MCU		        : bit;
signal DATA_OUT_JOYPORT         : std_logic_vector(15 downto 0);
signal DATA_EN_JOYPORT          : bit;
signal DTACK_OUT_IDEn	        : bit;
signal DTACK_MCUn	            : bit;
signal DTACK_OUT_BLTn           : bit;
signal ROM_6_In			        : bit;
signal ROM_5_In			        : bit;
signal ROM_4_In			        : bit;
signal ROM_3_In			        : bit;
signal ROM_2_In			        : bit;
signal ROM_1_In			        : bit;
signal ROM_0_In			        : bit;
signal SNDCS_In			        : bit;
signal DE_I				        : bit;
signal KHz_500_I                : bit;
signal RTCCS_In			        : bit;
signal JOY_RS_I			        : bit;
signal PAD0X_RS_I		        : bit;
signal PAD0Y_RS_I		        : bit;
signal PAD1X_RS_I		        : bit;
signal PAD1Y_RS_I		        : bit;
signal BUTTON_RS_I		        : bit;
signal SCC_In	                : bit;
signal VCS_I                    : bit;
signal MEM_CONFIG_RS_I          : bit;
signal LINE_OFFS_RS_I           : bit;
signal LINE_WIDTH_RS_I          : bit;
signal VIDEO_BASE_HI_RS_I       : bit;
signal VIDEO_BASE_MID_RS_I      : bit;
signal VIDEO_BASE_LOW_RS_I      : bit;
signal VIDEO_COUNT_HI_RS_I      : bit;
signal VIDEO_COUNT_MID_RS_I     : bit;
signal VIDEO_COUNT_LOW_RS_I     : bit;
signal R8006CS_I                : bit;
signal R8007CS_I                : bit;
signal FPUCS_I                  : bit;
signal RAM_In                   : bit;
signal R8007_REG                : std_logic_vector(7 downto 0);
signal SHMOD_ST_SHADOW_RS_I     : bit;
signal VMODE_SHADOW_RS_I        : bit;
signal VIDEO_BASE_HIWORD_RS_I   : bit;
signal VIDEO_BASE_LOWORD_RS_I   : bit;
signal VIDEO_COUNT_HIWORD_RS_I  : bit;
signal VIDEO_COUNT_LOWORD_RS_I  : bit;
begin
    ADR_IN_I <= ADR_IN when ADR_16MB_UP = '1' else x"FF" & ADR_IN(23 downto 1);

    DATA_EN <= DATA_EN_BLITTER or DATA_EN_MCU or DATA_EN_JOYPORT;
     -- x"FF" due to pull up resistors in original hardware.
    DATA_OUT <= DATA_OUT_BLITTER when DATA_EN_BLITTER = '1' else
                DATA_OUT_MCU when DATA_EN_MCU = '1' else
                DATA_OUT_JOYPORT when DATA_EN_JOYPORT = '1' else 
                x"FF" & DATA_IN(15 downto 8) when R8006CS_I = '1' and RWn_IN = '1' else
                x"FF" & R8007_REG when R8007CS_I = '1' and RWn_IN = '1' else 
                DATA_IN(15 downto 8) & R8007_REG when R8006CS_I = '1' and R8007CS_I = '1' and RWn_IN = '0' else (others => '0');

    -- System controls:
    KHz_500 <= KHz_500_I;
    KHz_500W <= KHz_500_I when TOK = '1' else '0';
    BERRn <= BERR_In;

    -- Register selections:
    R8006 <= R8006CS_I; -- Flacon's configuration register.
    FPUCS <= FPUCS_I; -- Floating point unit.
    VCS <= VCS_I; -- VIDEL video chip.

    -- TOS-ROMS:
	ROM_6n <= ROM_6_In;
	ROM_5n <= ROM_5_In;
	ROM_4n <= ROM_4_In;
	ROM_3n <= ROM_3_In;
	ROM_2n <= ROM_2_In;
	ROM_1n <= ROM_1_In;
	ROM_0n <= ROM_0_In;
	
	-- Soundchip:
	SNDCS 	<= '1' when SNDCS_In = '0' and ADR_IN_I(1) = '0' else '0';
	SNDIR 	<= '1' when RWn_IN = '0' and SNDCS_In = '0' else '0';

    -- Serial communication controller:
	SCCn <= SCC_In; -- SCC serial controller.
    SCCRD <= '1' when SCC_In = '0' and RWn_IN = '1' else '0';
    SCCWR <= '1' when SCC_In = '0' and RWn_IN = '0' else '0';

    -- Real time clock:
	RTCCSn 	<= RTCCS_In;
	RTC_RDn <= '0' when RWn_IN = '1' and LDS_INn = '0' and VMAn = '0' else '1';
	RTC_WRn <= '0' when RWn_IN = '0' and LDS_INn = '0' and VMAn = '0' else '1';
	
	-- Joyport:
	JOY_RHn 	<= '0' when JOY_RS_I = '1' and RWn_IN = '1' and UDS_INn = '0' else '1'; 
	JOY_RLn 	<= '0' when JOY_RS_I = '1' and RWn_IN = '1' and LDS_INn = '0' else '1'; 
	JOY_WL 	<= '1' when JOY_RS_I = '1' and RWn_IN = '0' and LDS_INn = '0' else '0'; 
	JOY_WEn 	<= '1' when JOY_RS_I = '1' and RWn_IN = '1' and LDS_INn = '0' else '0'; 
	BUTTONn 	<= '0' when BUTTON_RS_I = '1' else '1'; 	

    -- Feedback for the buscontrollers:
	DTACK_OUTn <= 	'0' when ROM_6_In = '0' 		            else
                    '0' when ROM_5_In = '0' 		            else
                    '0' when ROM_4_In = '0' 		            else
                    '0' when ROM_3_In = '0' 		            else
                    '0' when ROM_2_In = '0' 		            else
                    '0' when ROM_1_In = '0' 		            else
                    '0' when ROM_0_In = '0' 		            else
                    '0' when SNDCS_In = '0' 		            else
                    '0' when RTCCS_In = '0' 		            else
                    '0' when SCC_In = '0' and SCCWAITn = '1'    else
                    '0' when JOY_RS_I = '1' 		            else
                    '0' when PAD0X_RS_I = '1' 		            else
                    '0' when PAD0Y_RS_I = '1' 		            else
                    '0' when PAD1X_RS_I = '1' 		            else
                    '0' when PAD1Y_RS_I = '1' 		            else
                    '0' when BUTTON_RS_I = '1' 		            else
                    '0' when R8006CS_I = '1'                    else
                    '0' when R8007CS_I = '1'                    else
                    '0' when FPUCS_I = '1'                      else
                    '0' when VCS_I = '1'                        else
                    '0' when LINE_OFFS_RS_I = '1' 	            else
                    '0' when LINE_WIDTH_RS_I = '1' 	            else
                    '0' when MEM_CONFIG_RS_I = '1'		        else
                    '0' when VIDEO_BASE_HI_RS_I = '1' 	        else
                    '0' when VIDEO_BASE_MID_RS_I = '1' 	        else
                    '0' when VIDEO_BASE_LOW_RS_I = '1' 	        else
                    '0' when VIDEO_COUNT_HI_RS_I = '1' 	        else
                    '0' when VIDEO_COUNT_MID_RS_I = '1'         else
                    '0' when VIDEO_COUNT_LOW_RS_I = '1'         else
                    '0' when VIDEO_BASE_HIWORD_RS_I = '1'       else
                    '0' when VIDEO_BASE_LOWORD_RS_I = '1'       else
                    '0' when VIDEO_COUNT_HIWORD_RS_I = '1'      else
                    '0' when VIDEO_COUNT_LOWORD_RS_I = '1'      else
                    '0' when DTACK_OUT_IDEn = '0'               else
                    '0' when DTACK_MCUn = '0'                   else
                    '0' when DTACK_OUT_BLTn = '0'               else '1';

	SYSTEM_REG: process(CLK_32, R8007_REG)
    --    $FF8007|Falcon Bus Control          BIT . . 5 . . 2 . 0 |R/W
    --           |STe Bus Emulation (0 - on) ---------'     |   |
    --           |Blitter (0 - 8mhz, 1 - 16mhz) ------------'   |
    --           |68030 (0 - 8mhz, 1 - 16mhz) ------------------'
	begin
		if CLK_32 = '1' and CLK_32' event then
            if RESET = '1' then
                R8007_REG <= x"00";
			elsif R8007CS_I = '1' and RWn_IN = '0' then
                R8007_REG <= DATA_IN(7 downto 0);
            end if;
		end if;
		--
        CPU_CLKSEL_16_8n <= To_Bit(R8007_REG(0));
        BLT_CLKSEL_16_8n <= To_Bit(R8007_REG(2));
	end process SYSTEM_REG;

    I_AUXCLOCKS: CLOCKS
    port map(
        CLK                 => CLK_32,

        CLK_04              => CLK_8,
        CLK_08              => CLK_4,
        CLK_064             => KHz_500_I
    );

    I_BLITTER: BLITTER_TOP
        port map(
        CLK			        => CLK_BLT,
        RESET		        => RESET,
        AS_INn		        => AS_INn,
        AS_OUTn             => AS_OUTn,
        LDS_INn             => LDS_INn,
        LDS_OUTn	        => LDS_OUTn,
        UDS_INn		        => UDS_INn,
        UDS_OUTn    	    => UDS_OUTn,
        RWn_IN              => RWn_IN,
        RWn_OUT		        => RWn_OUT,
        DTACK_INn	        => DTACK_INn,
        DTACK_OUTn			=> DTACK_OUT_BLTn,
        BERRn		        => BERR_In,
        BMODE               => BMODE,
        FC_IN		        => FC_IN,
        FC_OUT		        => FC_OUT,
        BUSCTRL_EN	        => BUS_EN,
        INTn		        => BINTn,

        ADR_IN		        => ADR_IN_I,
        ADR_OUT		        => ADR_OUT,
        ADR_EN		        => ADR_EN,
        DATA_IN		        => DATA_IN,
        DATA_OUT	        => DATA_OUT_BLITTER,
        DATA_EN		        => DATA_EN_BLITTER,

        BGIn		        => BGIn,
        BRn			        => BRn,
        BGACK_INn	        => BGACK_INn,
        BGACK_OUTn	        => BGACK_OUTn,
        BGOn		        => BGOn
        );

	I_ADRDEC: ADRDEC
	port map(
		ADR 			        => ADR_IN_I,
		RWn 			        => RWn_IN,

		ROMSEL_FC_E0n           => ROMSEL_FC_E0n,

		LDSn 			        => LDS_INn,
		UDSn 			        => UDS_INn,
				
		ASn 			        => AS_INn,
		
		VPAn 			        => VPAn,
		VMAn 			        => VMAn,

		FC 				        => FC_IN,

		ROM_0n 			        => ROM_0_In,
		ROM_1n 			        => ROM_1_In,
		ROM_2n 			        => ROM_2_In,
		ROM_3n 			        => ROM_3_In,
		ROM_4n 			        => ROM_4_In,
		ROM_5n 			        => ROM_5_In,
		ROM_6n 			        => ROM_6_In,

		ACIACS 			        => N6850,
		MFPCSn 			        => MFPCSn,
		SNDCSn 			        => SNDCS_In,
		SCCn			        => SCC_In,
		SCCAB                   => SCCAB,
		RTCCSn			        => RTCCS_In,

		JOY_RS	 		        => JOY_RS_I,

		PAD0X_RS	 	        => PAD0X_RS_I,
		PAD0Y_RS 		        => PAD0Y_RS_I,
		PAD1X_RS 		        => PAD1X_RS_I,
		PAD1Y_RS 		        => PAD1Y_RS_I,

		BUTTON_RS 		        => BUTTON_RS_I,

        R8006CS                 => R8006CS_I,
        R8007CS                 => R8007CS_I,

        FPUCS                   => FPUCS_I,

        VCS                     => VCS_I,

        MEM_CONFIG_RS           => MEM_CONFIG_RS_I,
        LINE_OFFS_RS            => LINE_OFFS_RS_I,
        LINE_WIDTH_RS           => LINE_WIDTH_RS_I,

        VIDEO_BASE_HIWORD_RS 	=> VIDEO_BASE_HIWORD_RS_I,
        VIDEO_BASE_LOWORD_RS 	=> VIDEO_BASE_LOWORD_RS_I,
        VIDEO_COUNT_HIWORD_RS 	=> VIDEO_COUNT_HIWORD_RS_I,
        VIDEO_COUNT_LOWORD_RS 	=> VIDEO_COUNT_LOWORD_RS_I,
        VIDEO_BASE_HI_RS        => VIDEO_BASE_HI_RS_I,
        VIDEO_BASE_MID_RS       => VIDEO_BASE_MID_RS_I,
        VIDEO_BASE_LOW_RS       => VIDEO_BASE_LOW_RS_I,
        VIDEO_COUNT_HI_RS       => VIDEO_COUNT_HI_RS_I,
        VIDEO_COUNT_MID_RS      => VIDEO_COUNT_MID_RS_I,
        VIDEO_COUNT_LOW_RS      => VIDEO_COUNT_LOW_RS_I,

        SHMOD_ST_SHADOW_RS      => SHMOD_ST_SHADOW_RS_I,
        VMODE_SHADOW_RS         => VMODE_SHADOW_RS_I,

        RAMn 			        => RAM_In
	);

    I_MCU: MCU_TOP
	port map(  
        CLK                     => CLK_32,

		RESET                   => RESET,

		LDSn                    => LDS_INn,
		UDSn                    => UDS_INn,
		RWn                     => RWn_IN,

		ADR_IN                  => ADR_IN_I,
        
        RAMn                    => RAM_In,
        RAMH                    => RAMH,
        VREQ                    => VREQ,
        EVENn_ODD               => EVENn_ODD,

        RDAT                    => RDAT,
        WDAT                    => WDAT,
        VLD                     => VLD,

        VINTn                   => VINTn,
        HINTn                   => HINTn,

        VIDEO_BASE_HIWORD_RS 	=> VIDEO_BASE_HIWORD_RS_I,
        VIDEO_BASE_LOWORD_RS 	=> VIDEO_BASE_LOWORD_RS_I,
        VIDEO_COUNT_HIWORD_RS 	=> VIDEO_COUNT_HIWORD_RS_I,
        VIDEO_COUNT_LOWORD_RS 	=> VIDEO_COUNT_LOWORD_RS_I,
        VIDEO_BASE_HI_RS        => VIDEO_BASE_HI_RS_I,
        VIDEO_BASE_MID_RS       => VIDEO_BASE_MID_RS_I,
        VIDEO_BASE_LOW_RS       => VIDEO_BASE_LOW_RS_I,
        VIDEO_COUNT_HI_RS       => VIDEO_COUNT_HI_RS_I,
        VIDEO_COUNT_MID_RS      => VIDEO_COUNT_MID_RS_I,
        VIDEO_COUNT_LOW_RS      => VIDEO_COUNT_LOW_RS_I,

        SHMOD_ST_SHADOW_RS      => SHMOD_ST_SHADOW_RS_I,
        VMODE_SHADOW_RS         => VMODE_SHADOW_RS_I,
        MEM_CONFIG_RS           => MEM_CONFIG_RS_I,
        LINE_OFFS_RS            => LINE_OFFS_RS_I,
        LINE_WIDTH_RS           => LINE_WIDTH_RS_I,

        DTACKn                  => DTACK_MCUn,

        DATA_IN                 => DATA_IN,
        DATA_OUT                => DATA_OUT_MCU,
		DATA_EN                 => DATA_EN_MCU,

        BANK0_TYPE              => BANK0_TYPE,
        RAS1n                   => RAS1n,
        CAS1Hn                  => CAS1Hn,
        CAS1Ln                  => CAS1Ln,
        RAS0n                   => RAS0n,
        CAS0Hn                  => CAS0Hn,
        CAS0Ln                  => CAS0Ln,
        WEn                     => WEn,
        RAM_MADR                => RAM_ADR
	);

	I_INTERRUPT: INTERRUPTS
	port map(
		RESET 			    => RESET,
		CLK 			    => CLK_32,
		ADR_HI 			    => ADR_IN_I(19 downto 16),
		ADR_LO 			    => ADR_IN_I(3 downto 1),
		FC 				    => FC_IN,
		ASn 			    => AS_INn,
		EINT1n			    => EINT1n,
		EINT3n			    => EINT3n,
		EINT5n			    => EINT5n,
		EINT7n			    => EINT7n,
		MFPINTn 		    => MFPINTn,
		HINTn			    => HINTn,
		VINTn 			    => VINTn,
		AVECn 			    => AVECn,
		IACKn 			    => IACKn,
		SCCIACKn 			=> SCCIACKn,
		IPLn(2)			    => IPL2n,
		IPLn(1)			    => IPL1n,
 		IPLn(0)			    => IPL0n
	);

	I_ErrorHandler: ERRHANDLER
	port map(
		RESET 		        => RESET,
		CLK			        => CLK_32,
		ASn			        => AS_INn,
		BERRn		        => BERR_In
	);

	I_JOYPORT: JOYPORT_PADDLES
	port map(
        CLK 			    => CLK_32,
        KHz_500		        => KHz_500_I,
		RESET  			    => RESET,
		
		RWn				    => RWn_IN,
		DATA_IN 		    => DATA_IN(1 downto 0),
		DATA_OUT		    => DATA_OUT_JOYPORT,
		DATA_EN			    => DATA_EN_JOYPORT,

		PAD0X_RS 		    => PAD0X_RS_I,
		PAD0Y_RS 		    => PAD0Y_RS_I,
		PAD1X_RS 		    => PAD1X_RS_I,
		PAD1Y_RS 		    => PAD1Y_RS_I,
		
		PAD0X_INHn 		    => PAD0Xn,
		PAD0Y_INHn 		    => PAD0Yn,
		PAD1X_INHn 		    => PAD1Xn,
		PAD1Y_INHn 		    => PAD1Yn,
		PADRSTn 		    => PADRSTn
	);

    I_IDE: WF_IDE
    port map(
        RESET               => RESET,

        ADR                 => ADR_IN_I(31 downto 4),

        ASn                 => AS_INn,
        LDSn                => LDS_INn,
        RWn                 => RWn_IN,
        DTACKn              => DTACK_OUT_IDEn,
        
        -- Interrupt:
        -- HDINTn           =>, -- Not used so far.
        
        -- IDE section:
        IDE_INTRQ           => '0', -- Not used.
        IDE_IORDY           => '1', -- Not used.
        -- PDIAG            =>, -- Not used so far.
        -- DASP             =>, -- See pinout above.
        -- DMARQ            =>, -- Not used so far.
        -- DMACKn           =>, -- See pinout above.
        -- IDE_RESn         =>, -- Not used.
        CS0n                => IDE_RS0n,
        CS1n                => IDE_RS1n,
        IORDn               => IDE_IORDn,
        IOWRn               => IDE_IOWRn,
        
        IDE_D_EN_INn        => IDE_D_EN_INn,
        IDE_D_EN_OUTn       => IDE_D_EN_OUTn
      );

end STRUCTURE;
