----------------------------------------------------------------------
----                                                              ----
---- ATARI MCU compatible IP Core					              ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- Memory management controller with all features to reach      ----
---- ATARI STE compatibility.                                     ----
---- This controller meets the requirements for SD-RAMs. See the  ----
---- top level file for more information. This controller is in   ----
---- comparision to the original ATARI controller enhanced by the ----
---- 8192KWords SD-RAM size. This is selected for the respective  ----
---- memory bank by the bits 3 downto 0 of the memory config      ----
---- register. See the coding for the BANK0_TYPE, the BANK_SWITCH ----
---- and the DTACKn in this file for more information.            ----
----                                                              ----
---- Cotrol file for the MCU.                                     ----
----                                                              ----
----                                                              ----
---- To Do:                                                       ----
---- -                                                            ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright � 2009-2010 Wolfgang Foerster Inventronik GmbH.    ----
---- All rights reserved. No portion of this sourcecode may be    ----
---- reproduced or transmitted in any form by any means, whether  ----
---- by electronic, mechanical, photocopying, recording or        ----
---- otherwise, without my written permission.                    ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K6A  2006/06/03 WF
--   Initial Release.

library work;
use work.COMBEL_PKG.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

entity MCU_CTRL is
	port (  CLK				: in bit;
			RESET 			: in bit;

			LDSn, UDSn, RWn	: in bit; -- Bus control signals.

			M_ADR			: in bit_vector(31 downto 1); -- Non multiplexed DRAM addresses.
			
            RAMn			: in bit;
            RAMH			: out bit;
            VREQ            : in bit; 

			MEM_CONFIG_RS	: in bit; -- Memory config register control.
			MCU_PHASE		: out MCU_PHASE_TYPE;
			
            BANK0_TYPE      : out BANKTYPE; -- Memory type indicator.

			VINTn			: in bit; -- Vertical rollback.

			RAS0n			: out bit; -- Memory bank 1 row address strobe.
			CAS0Hn			: out bit; -- Memory bank 1 column address strobe.
			CAS0Ln			: out bit; -- Memory bank 1 column address strobe.

			WEn				: out bit; -- Memory write control, low active.

			RAS1n			: out bit; -- Memory bank 2 row address strobe.
			CAS1Hn			: out bit; -- Memory bank 2 column address strobe.
			CAS1Ln			: out bit; -- Memory bank 2 column address strobe.

			RDAT 			: out bit; -- Buffer control.
			WDAT 			: out bit; -- Buffer control.
			VLD 			: out bit; -- Buffer control.
			
			REF_EN			: out bit; -- Refresh counter enable.
			VIDEO_CNT_EN	: out bit; -- Video control.
			VIDEO_CNT_LOAD	: out bit; -- Video control.

			DTACKn			: out bit; -- Data acknowledge signal.

			DATA_IN			: in std_logic_vector(7 downto 0);
			DATA_OUT		: out bit_vector(7 downto 0);
			DATA_EN			: out bit
	);
end MCU_CTRL;

architecture BEHAVIOR of MCU_CTRL is
type BANKS is (BANK1, BANK0);
type MATRIX_ELEMENTS is array (1 to 10, 1 to 8) of bit;
constant TIME_MATRIX : MATRIX_ELEMENTS := 
	(('0','1','1','1','0','1','1','1'),		-- RASn.
	 ('1','0','0','1','1','0','0','1'),		-- CASHn.
	 ('1','0','0','1','1','0','0','1'),		-- CASLn.
	 ('1','0','0','1','1','1','1','1'),		-- WEn.
     ('0','0','0','0','1','1','1','1'),		-- RDATn.
     ('0','0','0','0','1','1','1','1'),		-- WDATn.
	 ('0','0','1','1','0','0','0','0'),		-- VLD.
	 ('1','0','1','0','0','0','0','0'),		-- REF_EN.
	 ('0','0','0','1','0','0','0','1'),		-- VIDEO_CNT_EN.
     ('1','0','0','0','1','1','1','1'));	-- DTACK_MASKn.

signal MCU_PHASE_I			: MCU_PHASE_TYPE;
signal BANK0_TYPE_I 		: BANKTYPE; -- Memory type indicator.
signal MEMCONFIG			: std_logic_vector(7 downto 0);
signal SLICE_NUMBER			: integer range 1 to 8;
signal BANK_SWITCH			: BANKS;
signal M_ADR_I 				: bit_vector(31 downto 0);

signal RASn					: bit;
signal CASLn				: bit; 
signal CASHn				: bit;

signal DTACK_MASK_In 		: bit;

signal VIDEO_CNT_EN_I		: bit;
begin

	MEMCONFIG_REG: process
	begin
		-- The MEMCONFIG must start up with x"0A" indicating virtual 4MB RAM.
		-- Otherwise the RAM test routine will hang due to no DTACKn signal.
		-- The value of x"0A" is written immediately after system startup by
		-- the CPU. This register may not have a clear via RESET because the
		-- operating system does not initialise it after an interrupt like
		-- monochrome detect (level 15).
		wait until CLK = '1' and CLK' event;
		if MEM_CONFIG_RS = '1' and RWn = '0' then
			MEMCONFIG <= DATA_IN; -- Write to register.
		end if;
	end process MEMCONFIG_REG;

	-- Read memory config register:
	DATA_OUT <= To_BitVector(MEMCONFIG) when MEM_CONFIG_RS = '1' and RWn = '1' else (others => '0');
	DATA_EN <= MEM_CONFIG_RS and RWn;

	-- Address control:
	M_ADR_I <= M_ADR & '0';

	MCU_PHASE <= MCU_PHASE_I;

	-- Bits 7, 6, 3 and 2 of the MEMCONFIG select the bank 0 and the bits
	-- 5, 4, 1 and 0 are for the selection of bank 1.
	BANK0_TYPE <= BANK0_TYPE_I;
    with MEMCONFIG(7 downto 6) & MEMCONFIG(3 downto 2) select
		BANK0_TYPE_I <= K2048x16 when x"7",
                        K1024x16 when x"6",
                        K512x16 when x"5",
                        K256x16 when x"4",
                        K8192 when x"3",
                        K2048 when x"2",
                        K512  when x"1",
                        K128 when others;

	-- SD-RAM control signals:
	RAS0n <= RASn when MCU_PHASE_I = VIDEO else -- Video is 32 bit access.
             RASn when BANK_SWITCH = BANK0 else '1';
	RAS1n <= RASn when MCU_PHASE_I = VIDEO else -- Video is 32 bit access.
             RASn when BANK_SWITCH = BANK1 else '1';
	CAS0Ln <= CASLn when MCU_PHASE_I = VIDEO else -- Video is 32 bit access.
              CASLn when BANK_SWITCH = BANK0 else '1';
	CAS1Ln <= CASLn when MCU_PHASE_I = VIDEO else -- Video is 32 bit access.
              CASLn when BANK_SWITCH = BANK1 else '1';
	CAS0Hn <= CASHn when MCU_PHASE_I = VIDEO else -- Video is 32 bit access.
              CASHn when BANK_SWITCH = BANK0 else '1';
	CAS1Hn <= CASHn when MCU_PHASE_I = VIDEO else -- Video is 32 bit access.
              CASHn when BANK_SWITCH = BANK1 else '1';

	-- Shifter stuff:
	VIDEO_CNT_LOAD <= '1' when VINTn = '0' else '0';

	-- Bus, memory and system control signals:
	RASn	<= TIME_MATRIX(1, SLICE_NUMBER) when MCU_PHASE_I = REFRESH or MCU_PHASE_I = RAM else
			   TIME_MATRIX(1, SLICE_NUMBER) when MCU_PHASE_I = VIDEO else '1';
	CASHn 	<= TIME_MATRIX(2, SLICE_NUMBER) when MCU_PHASE_I = RAM and UDSn = '0' else
			   TIME_MATRIX(2, SLICE_NUMBER) when MCU_PHASE_I = VIDEO else '1';
	CASLn	<= TIME_MATRIX(3, SLICE_NUMBER) when MCU_PHASE_I = RAM and LDSn = '0' else
			   TIME_MATRIX(3, SLICE_NUMBER) when MCU_PHASE_I = VIDEO else '1';
	WEn		<= TIME_MATRIX(4, SLICE_NUMBER) when MCU_PHASE_I = RAM and RWn = '0' else '1'; -- Write is valid in RAM mode.

	RDAT 	<= not TIME_MATRIX(5, SLICE_NUMBER) when MCU_PHASE_I = RAM  and RWn = '1' else '0'; -- Read to bus only in RAM mode.
	WDAT 	<= not TIME_MATRIX(6, SLICE_NUMBER) when MCU_PHASE_I = RAM  and RWn = '0' else '0';
    VLD 	<= not TIME_MATRIX(7, SLICE_NUMBER) when MCU_PHASE_I = VIDEO  else '0';

	REF_EN		<= TIME_MATRIX(8, SLICE_NUMBER) when MCU_PHASE_I = REFRESH else '0';
	
	VIDEO_CNT_EN_I 	<= TIME_MATRIX(9, SLICE_NUMBER) when MCU_PHASE_I = VIDEO else '0'; -- Video data.

	VIDEO_CNT_EN <= VIDEO_CNT_EN_I;

    DTACK_MASK_In	<= TIME_MATRIX(10, SLICE_NUMBER) when MCU_PHASE_I = RAM else '1'; -- RAM, SHIFTER register access.

	-- RAM bank select logic. It is dependant on the equiped memory. The BANK0 is the lower one in the
    -- ATARI's address space. The first 4 memory configurations are legacy for compatibility reasons.
    -- the next four modes are intended for use of DDR(2) RAMs with 16Mx16, 32Mx16, 64Mx16 and 128Mx16.
    -- In this way a maximum RAM space results in 512MByte if both banks are equiped with 64Mx16.
	BANK_SWITCH <= BANK0 when BANK0_TYPE_I = K128  and M_ADR_I < x"00020000" else -- 64Kwords in bank 0.
                   BANK0 when BANK0_TYPE_I = K512  and M_ADR_I < x"00080000" else -- 256Kwords in bank 0.
                   BANK0 when BANK0_TYPE_I = K2048 and M_ADR_I < x"00200000" else -- 1024Kwords in bank 0.
                   BANK0 when BANK0_TYPE_I = K8192 and M_ADR_I < x"00800000" else -- 4096Kwords in bank 0.
				   BANK0 when BANK0_TYPE_I = K256x16 and M_ADR_I < x"01000000" else -- 256MBitx16 in bank 0.
				   BANK0 when BANK0_TYPE_I = K512x16 and M_ADR_I < x"02000000" else -- 512MBitx16 in bank 0.
				   BANK0 when BANK0_TYPE_I = K1024x16 and M_ADR_I < x"04000000" else -- 1024MBitx16 in bank 0.
				   BANK0 when BANK0_TYPE_I = K2048x16 and M_ADR_I < x"08000000" else BANK1; -- 2048MBitx16 in bank 0, 

    RAMH <= '1' when BANK_SWITCH = BANK1 else '0';

    DTACKn <= DTACK_MASK_In;

	MCU_PHASE_SWITCH: process
	-- AD MCU_PHASE_TYPES: SHIFTER is foreseen to  access the shifter
	-- registers; VIDEO transfers video data from RAM to shifter; RAM 
	-- is the CPU access.
	-- The REFRESH cycle is foreseen to hold the data in the SD-RAMs.
	-- This process controls the type of data transfer in the second period
	-- of the MCU cycle (250ns ... 500ns). While the first half of the MCU
	-- cycle is reserved for data transfer to the shifter, the second one
	-- shares data transfer between CPU and RAM and is foreseen for the 
	-- RAM REFRESH process and the data transfer to the shifter registers 
	-- (MCU_PHASE = SHIFTER).
    -- The TMP variable controls the video access every second phase in
	-- case of a fast clocked MCU (double clock rate than the original MCU).
	begin
		wait until CLK = '1' and CLK' event;
		if RESET = '1' then
            MCU_PHASE_I <= REFRESH; -- REFRESH during reset keeps data alive.
        elsif SLICE_NUMBER = 4 then
            if VREQ = '1' then
                MCU_PHASE_I <= VIDEO; -- Video data out.
            else
                MCU_PHASE_I <= IDLE; -- Do nothing, wait for one of the data cycles.
            end if;
        elsif SLICE_NUMBER = 8 then
            if VREQ = '1' then -- Video access has priority over RAM access.
                MCU_PHASE_I <= VIDEO; -- Video data out.
            elsif RAMn = '0' then -- CPU access or coprocessor access (via bus arbitration).
                MCU_PHASE_I <= RAM;
            else -- REFRESH, if no data transfer is required.
                MCU_PHASE_I <= REFRESH;
            end if;
        end if;
	end process MCU_PHASE_SWITCH;

	TIME_SLICES: process
	-- This counter may not have a reset control because it has to
	-- Produce refresh timing during reset. The process must operate
	-- on the positive clock edge to synchronize correctly with the
	-- GLUE DMA part.
    -- Important enhancement! The first access, the RAMn signal
	-- synchronizes the MCU clock with the phase of the CPU clock. 
	-- This is very important for correct system startup concerning
	-- the DTACKn detection in the CPU and it's waitstate generation
	-- A wrong clock phase terminates the CPU access too early 
	-- resulting in malfunction. Generally spoken a CPU busphase with
	-- an equal number (S0, S2...) must match an odd MCU SLICE_NUMBER
	-- (1, 3...)!
	variable TIME_SLICE_CNT : std_logic_vector(2 downto 0);
    variable LOCK   : boolean;
	begin
		wait until CLK = '1' and CLK' event;
        if RESET = '1' then
            LOCK := false;
        elsif RAMn = '0' and LOCK = false then -- Sync once!
            LOCK := true;
            TIME_SLICE_CNT := "111";
        else
            TIME_SLICE_CNT := TIME_SLICE_CNT + '1';
        end if;
        --
        SLICE_NUMBER <= conv_integer(TIME_SLICE_CNT) + 1;
	end process TIME_SLICES;
end architecture BEHAVIOR;
