----------------------------------------------------------------------
----                                                              ----
---- ATARI MCU compatible IP Core					              ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- Memory management controller with all features to reach      ----
---- ATARI STE compatibility.                                     ----
----                                                              ----
---- The video access is done by the counter in this file.        ----
----                                                              ----
----                                                              ----
---- To Do:                                                       ----
---- -                                                            ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright � 2009-2010 Wolfgang Foerster Inventronik GmbH.    ----
---- All rights reserved. No portion of this sourcecode may be    ----
---- reproduced or transmitted in any form by any means, whether  ----
---- by electronic, mechanical, photocopying, recording or        ----
---- otherwise, without my written permission.                    ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K6A  2006/06/03 WF
--   Initial Release.
-- Revision 2K7A  2007/01/02 WF
--   Changes to the clock system and related
--   hardware as sound or video control.
-- Revision 2K8A  2008/07/14 WF
--   Minor changes.
-- Revision 2K9A  2008/12/24 WF
--   Introduced multisync compatibility modes (s. VIDEO_CNT).
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity MCU_VIDEO_COUNTER is
port (  CLK		                : in bit;
		RESET 	                : in bit;
		RWn		                : in bit; -- Read write control.

        VIDEO_BASE_HIWORD_RS	: in bit; -- Hi portion of the 32 bit wide video base.
        VIDEO_BASE_LOWORD_RS	: in bit; -- Lo portion of the 32 bit wide video base.
        VIDEO_COUNT_HIWORD_RS	: in bit; -- Hi portion of the 32 bit wide video counter.
        VIDEO_COUNT_LOWORD_RS	: in bit; -- Lo portion of the 32 bit wide video counter.
		
		VIDEO_BASE_HI_RS	    : in bit; -- Register control signal.
		VIDEO_BASE_MID_RS	    : in bit; -- Register control signal.
		VIDEO_BASE_LOW_RS	    : in bit; -- Register control signal.

		VIDEO_COUNT_HI_RS	    : in bit; -- Register control signal.
		VIDEO_COUNT_MID_RS	    : in bit; -- Register control signal.
		VIDEO_COUNT_LOW_RS	    : in bit; -- Register control signal.
		
        SHMOD_ST_SHADOW_RS      : in bit; -- Shadow acces for the ST shift mode register located in the Videl.
        VMODE_SHADOW_RS         : in bit; -- Shadow acces for the video mode register located in the Videl.

        VINTn			        : in bit; -- Vertical interrupt.
        HINTn			        : in bit; -- Horizontal interrupt.

		EVENn_ODD               : in bit; -- Interlaced video frame indicator.
		VIDEO_COUNT_EN		    : in bit; -- Counter enable.
		VIDEO_COUNT_LOAD	    : in bit; -- Load control.

        LINE_OFFS_RS		    : in bit; -- Select for the STEs linewidth register.
        LINE_WIDTH_RS		    : in bit; -- Select for the STEs linewidth register.
		
		VIDEO_ADR_OUT		    : out bit_vector(31 downto 1);
		
		DATA_IN			        : in std_logic_vector(15 downto 0);
		DATA_OUT		        : out std_logic_vector(15 downto 0);
		DATA_EN			        : out bit
      );
end MCU_VIDEO_COUNTER;

architecture BEHAVIOR of MCU_VIDEO_COUNTER is
signal VIDEO_BASE	    : std_logic_vector(31 downto 0);
signal VIDEO_ADR		: std_logic_vector(31 downto 1);
signal LINEOFFSET	    : std_logic_vector(8 downto 0);
signal LINEWIDTH	    : std_logic_vector(9 downto 0);
signal LINE_RELOAD      : bit;
signal INTERLACED_EN    : bit;
signal DOUBLELINE_EN    : bit;
begin
	VIDEO_REGS: process
	variable LOCK : boolean;
	begin
		wait until CLK = '1' and CLK' event;
        if RESET = '1' then
            VIDEO_BASE <= (others => '0');
            VIDEO_ADR <= (others => '0');
            LINEOFFSET <= (others => '0');
            LINEWIDTH <= (others => '0');
            LOCK := false;
        end if;
        --
        -- Due to this placement in the code, the write access to the video
        -- counter via the data bus has priority over VIDEO_COUNT_LOAD
        -- and VIDEO_COUNT_EN.
        if VIDEO_BASE_HIWORD_RS = '1' and RWn = '0' then -- x"FFFF8212"
            VIDEO_BASE(31 downto 16) <= DATA_IN;
            VIDEO_ADR(7 downto 1) <= (others => '0'); -- Clear video address low byte.
        end if;
        --
        if VIDEO_BASE_LOWORD_RS = '1' and RWn = '0' then -- x"FFFF8214"
            VIDEO_BASE(15 downto 0) <= DATA_IN;
            VIDEO_ADR(7 downto 1) <= (others => '0'); -- Clear video address low byte.
        end if;
        --
        if VIDEO_BASE_HI_RS = '1' and RWn = '0' then -- x"FFFF8201".
            VIDEO_BASE(31 downto 24) <= x"00"; -- Clear upper portion for compatibility reasons.
            VIDEO_BASE(23 downto 16) <= DATA_IN(7 downto 0);
            VIDEO_ADR(7 downto 1) <= (others => '0'); -- Clear video address low byte.
        end if;
        --
        if VIDEO_BASE_MID_RS = '1' and RWn = '0' then -- x"FFFF8203".
            VIDEO_BASE(15 downto 8) <= DATA_IN(7 downto 0);
            VIDEO_ADR(7 downto 1) <= (others => '0'); -- Clear video address low byte.
        end if;
        --
        if VIDEO_BASE_LOW_RS = '1' and RWn = '0' then -- x"FFFF820D".
            VIDEO_BASE(7 downto 0) <= DATA_IN(7 downto 0);
        end if;
        --
        if VIDEO_COUNT_HIWORD_RS = '1' and RWn = '0' then -- x"FFFF8216"
            VIDEO_ADR(31 downto 16) <= DATA_IN;
        end if;
        --
        if VIDEO_COUNT_LOWORD_RS = '1' and RWn = '0' then -- x"FFFF8218"
            VIDEO_ADR(15 downto 1) <= DATA_IN(15 downto 1);
        end if;
        --
        if VIDEO_COUNT_HI_RS = '1' and RWn = '0' then -- x"FFFF8205".
            VIDEO_ADR(31 downto 24) <= x"00"; -- Clear upper portion for compatibility reasons.
            VIDEO_ADR(23 downto 16) <= DATA_IN(7 downto 0);
        end if;
        --
        if VIDEO_COUNT_MID_RS = '1' and RWn = '0' then -- x"FFFF8207".
            VIDEO_ADR(15 downto 8) <= DATA_IN(7 downto 0);
        end if;
        --
        if VIDEO_COUNT_LOW_RS = '1' and RWn = '0' then -- x"FFFF8209".
            VIDEO_ADR(7 downto 1) <= DATA_IN(7 downto 1);
        end if;
        --
        if LINE_OFFS_RS = '1' and RWn = '0' then -- x"FFFF820E - FFFF820F"
            LINEOFFSET <= DATA_IN(8 downto 0);
        end if;
        --
        if LINE_WIDTH_RS = '1' and RWn = '0' then -- x"FFFF8210 - FFFF8211"
            LINEWIDTH <= DATA_IN(9 downto 0);
        end if;
        --
        if SHMOD_ST_SHADOW_RS = '1' and RWn = '0' then -- Shadow for x"FFFF8260"
            case DATA_IN(9 downto 8) is -- Change Linewidth.
                when "11" | "01" | "00" => LINEWIDTH <= "0010010000"; -- x"50"
                when others => LINEWIDTH <= "0000101000"; -- x"28"
            end case;
        end if;
        --
        if VMODE_SHADOW_RS = '1' and RWn = '0' then -- x"FFFF8210 - FFFF8211"
            INTERLACED_EN <= To_Bit(DATA_IN(1));
            DOUBLELINE_EN <= To_Bit(DATA_IN(0));
        end if;
        --
        if VIDEO_COUNT_LOAD = '1' then -- Load has priority over count_en.
             -- Begin the first line with n bytes delayed.
            VIDEO_ADR <= VIDEO_BASE(31 downto 1) + LINEOFFSET;
        end if;
        --
        if VIDEO_COUNT_EN = '1' and LINE_RELOAD = '1' then
            VIDEO_ADR <= VIDEO_ADR - LINEWIDTH + '1'; -- Rescan one line.
        elsif VIDEO_COUNT_EN = '1' and INTERLACED_EN = '1' then
            VIDEO_ADR <= VIDEO_ADR + LINEWIDTH;
            LOCK := false;
        elsif VIDEO_COUNT_EN = '1' then
            VIDEO_ADR <= VIDEO_ADR + '1';
            LOCK := false;
        end if;
        --
        if HINTn = '0' and LOCK = false and INTERLACED_EN = '1' and EVENn_ODD = '0' then
            VIDEO_ADR <= VIDEO_ADR + LINEOFFSET;
            LOCK := true;
        elsif HINTn = '0' and LOCK = false and INTERLACED_EN = '1' then
            VIDEO_ADR <= VIDEO_ADR + LINEOFFSET + LINEWIDTH;
            LOCK := true;
        elsif HINTn = '0' and LOCK = false then
            VIDEO_ADR <= VIDEO_ADR + LINEOFFSET;
            LOCK := true;
        end if;
	end process VIDEO_REGS;
	
	P_LINEDOUBLING: process
	variable SECOND_LINE	: bit;
	variable VIDEO_WORD_CNT	: std_logic_vector(6 downto 0);
	begin
		wait until CLK = '1' and CLK' event;
		--
        if HINTn = '0' or DOUBLELINE_EN = '0' then
			SECOND_LINE := '0';
			VIDEO_WORD_CNT := "0000000";
		elsif VIDEO_COUNT_EN = '1' and VIDEO_WORD_CNT < "1001111" then
			VIDEO_WORD_CNT := VIDEO_WORD_CNT + '1';
		elsif VIDEO_COUNT_EN = '1' and SECOND_LINE = '0' then
			VIDEO_WORD_CNT := "0000000";
			SECOND_LINE := '1';
		elsif VIDEO_COUNT_EN = '1' then
			VIDEO_WORD_CNT := "0000000";
			SECOND_LINE := '0';
		end if;
		--
		if VIDEO_WORD_CNT = (LINEWIDTH - '1') and SECOND_LINE = '0' then
			LINE_RELOAD <= '1';
		else
			LINE_RELOAD <= '0';
		end if;
	end process P_LINEDOUBLING;

	-- Read registers.
	DATA_OUT <= VIDEO_BASE(31 downto 16) when VIDEO_BASE_HIWORD_RS = '1' and RWn = '1' else
                VIDEO_BASE(15 downto 0) when VIDEO_BASE_LOWORD_RS = '1' and RWn = '1' else
                x"00" & VIDEO_BASE(23 downto 16) when VIDEO_BASE_HI_RS = '1' and RWn = '1' else
				x"00" & VIDEO_BASE(15 downto 8) when VIDEO_BASE_MID_RS = '1' and RWn = '1' else
				x"00" & VIDEO_BASE(7 downto 0) when VIDEO_BASE_LOW_RS = '1' and RWn = '1' else
				VIDEO_ADR(31 downto 16) when VIDEO_COUNT_HIWORD_RS = '1' and RWn = '1' else
				VIDEO_ADR(15 downto 1) & '0' when VIDEO_COUNT_LOWORD_RS = '1' and RWn = '1' else
				x"00" & VIDEO_ADR(23 downto 16) when VIDEO_COUNT_HI_RS = '1' and RWn = '1' else
				x"00" & VIDEO_ADR(15 downto 8) when VIDEO_COUNT_MID_RS = '1' and RWn = '1' else
				x"00" & VIDEO_ADR(7 downto 1) & '0' when VIDEO_COUNT_LOW_RS = '1' and RWn = '1' else 
				"0000000" & LINEOFFSET when LINE_OFFS_RS = '1' and RWn = '1' else
				"000000" & LINEWIDTH when LINE_WIDTH_RS = '1' and RWn = '1' else (others => '0');

	DATA_EN <= 	(VIDEO_BASE_HIWORD_RS or VIDEO_BASE_LOWORD_RS or VIDEO_COUNT_HIWORD_RS or VIDEO_COUNT_LOWORD_RS or
				 LINE_WIDTH_RS or LINE_OFFS_RS or VIDEO_BASE_HI_RS or VIDEO_BASE_MID_RS or VIDEO_BASE_LOW_RS or
				 VIDEO_COUNT_HI_RS or VIDEO_COUNT_MID_RS or VIDEO_COUNT_LOW_RS) and RWn;

	VIDEO_ADR_OUT <= To_BitVector(VIDEO_ADR);
end architecture BEHAVIOR;
