------------------------------------------------------------------------
----                                                                ----
----  Atari Falcon compatible direct memory access coprocessor.     ----
----  This file is part of the SUSKA-ATARI clone project.           ----
----                                                                ----
----  Author: Wolfgang Foerster                                     ----
----          support@inventronk.de                                 ----
----          www.inventronik.de                                    ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright (C) 2012 Inventronik GmbH                            ----
----                    Finkenstrasse 48                            ----
----                    70199 Stuttgart                             ----
----                    wf@inventronk.de                            ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Description:                                                   ----
----                                                                ----
------------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K12A  20120620 WF
--   Initial Release.
-- 

library ieee;
use ieee.std_logic_1164.all;

package FDMA_PKG is
    component FDMA_CTRL
        port (
            CLK                 : in bit;
            RESET               : in bit;

            -- Bus control:
            BMODE               : in bit;
            FC_OUT              : out bit_vector(2 downto 0);
            AS_INn              : in bit;
            AS_OUTn             : out bit;
            ADR_EN_ACSI         : out bit;
            ADR_EN_REPLAY       : out bit;
            ADR_EN_CAPTURE      : out bit;
            UDSn                : out bit;
            LDSn                : out bit;
            RWn                 : out bit;
            BERRn               : in bit; -- Bus error.
            DTACKn              : in bit;
            DATA_EN_ACSI        : out bit; -- Switch to connect the ACSI FIFO data to the bus.
            DATA_EN_CAPTURE     : out bit; -- Switch to connect the capture FIFO data to the bus.
            
            -- Bus arbitration:
            BRn                 : out bit;
            BGIn                : in bit;
            BGOn                : out bit;
            BGACK_INn           : in bit;
            BGACK_OUTn          : out bit;

            -- Control signals for ACSI:
            ACSI_FIFO_CLRn      : out bit; -- Invalidate the FIFO entries.
            ACSI_FIFO_WR        : out bit;
            ACSI_FIFO_RD        : out bit;
            ACSI_FIFO_FULL      : in bit;
            ACSI_FIFO_LOW       : in bit;
            ACSI_FIFO_EMPTY     : in bit;

            -- Control signals for the PCM out:
            REPLAY_FIFO_CLRn    : out bit; -- Invalidate the FIFO entries.
            REPLAY_FIFO_WR      : out bit;
            REPLAY_FIFO_RD      : out bit;
            REPLAY_FIFO_FULL    : in bit;
            REPLAY_FIFO_LOW     : in bit;

            -- Control signals for the PCM in:
            CAPTURE_FIFO_CLRn   : out bit; -- Invalidate the FIFO entries.
            CAPTURE_FIFO_WR     : out bit;
            CAPTURE_FIFO_RD     : out bit;
            CAPTURE_FIFO_LOW    : in bit;
            CAPTURE_FIFO_EMPTY  : in bit;

            -- Handshaking:
            HDRQ                : in bit; -- Data request from the sound device.
            HD_ACKn             : out bit;
            FDCRQ               : in bit; -- Data request from the sound device.
            FDCS_DMA_ACCn       : out bit;      
            ACSI_DATA_REQ       : out bit; -- Status register stuff.
            REPLAY_DATAREQ      : in bit; -- Data request from the sound device.
            REPLAY_DATACK       : out bit; -- Data request from the sound device.
            CAPTURE_DATAREQ     : in bit; -- Data request from the sound device.
            CAPTURE_DATACK      : out bit; -- Data request from the sound device.

            -- Counter controls:
            SECTOR_CNT_EN       : out bit;
            DMA_FRAME_CNT_EN    : out bit;
            RP_FRAME_CNT_EN     : out bit;
            CA_FRAME_CNT_EN     : out bit;

            -- Control signals for the ACSI multiplexer:
            CD_HIBUF_EN         : out bit; -- Writes ACSI_BUF_HI.
            CD_RD_HIn           : out bit; -- Reads high FIFO byte to CD.
            CD_RD_LOWn          : out bit; -- Reads low FIFO byte to CD.

            -- Other controls:
            DMA_EN              : in bit;
            PCM_REPLAY          : in bit;
            PCM_CAPTURE         : in bit;
            
            DMA_RWn             : in bit; -- FIFO direction '1' is peripherals to RAM.
            DMA_SRC_SEL         : in bit_vector(1 downto 0)
        );
    end component;

    component DMA_FIFOs
        generic(
            FIFO_WIDTH          : integer := 16; -- Number of bits per register.
            ACSI_FIFO_DEPTH     : integer := 16; -- Number of registers.
            REPLAY_FIFO_DEPTH   : integer := 16; -- Number of registers.
            CAPTURE_FIFO_DEPTH  : integer := 16 -- Number of registers.
            );
        port (
            RESET               : in bit;
            CLK                 : in bit;

            DATA_IN             : in std_logic_vector(FIFO_WIDTH -1 downto 0);
            DATA_OUT            : out std_logic_vector(FIFO_WIDTH -1 downto 0);

            CD_DATA_IN          : in std_logic_vector (7 downto 0);
            CD_DATA_OUT         : out std_logic_vector (7 downto 0);

            CAPTURE_DATA_IN     : in bit_vector (15 downto 0);
            REPLAY_DATA_OUT     : out bit_vector (15 downto 0);

            DMA_RWn             : in bit; -- FIFO direction '1' is peripherals to RAM.
            
            -- Control signals for the ACSI multiplexer:
            ACSI_DATA_EN        : in bit; -- Switch to connect the ACSI FIFO data to the bus.
            CD_HIBUF_EN         : in bit; -- Writes ACSI_BUF_HI.
            CD_RD_HIn           : in bit; -- Reads high FIFO byte to CD.
            CD_RD_LOWn          : in bit;  -- Reads low FIFO byte to CD.
            CAPTURE_DATA_EN     : in bit;  -- Switch to connect the PCM capture FIFO data to the bus.

            -- Control signals for the ACSI bus:
            ACSI_FIFO_CLRn      : in bit; -- Invalidate the FIFO entries.
            ACSI_FIFO_WR        : in bit;
            ACSI_FIFO_RD        : in bit;
            ACSI_FIFO_FULL      : out bit;
            ACSI_FIFO_LOW       : out bit;
            ACSI_FIFO_EMPTY     : out bit;
            ACSI_FIFO_ERROR     : out bit;

            -- Control signals for the PCM out:
            REPLAY_FIFO_CLRn    : in bit; -- Invalidate the FIFO entries.
            REPLAY_FIFO_WR      : in bit;
            REPLAY_FIFO_RD      : in bit;
            REPLAY_FIFO_FULL    : out bit;
            REPLAY_FIFO_LOW     : out bit;
            REPLAY_FIFO_EMPTY   : out bit;

            -- Control signals for the PCM in:
            CAPTURE_FIFO_CLRn   : in bit; -- Invalidate the FIFO entries.
            CAPTURE_FIFO_WR     : in bit;
            CAPTURE_FIFO_RD     : in bit;
            CAPTURE_FIFO_FULL   : out bit;
            CAPTURE_FIFO_LOW    : out bit;
            CAPTURE_FIFO_EMPTY  : out bit
        );
    end component;
    
    component FDMA_ACSI_REGs
        port(
            CLK                 : in bit;
            RESET               : in bit;

            FC                  : in bit_vector(2 downto 0); -- Processor function codes.
            ADR_IN              : in bit_vector(31 downto 1); --Adress inputs.
            LDSn                : in bit; -- Lower data strobe; not used so far.
            UDSn                : in bit; -- Upper data strobe.
            ASn                 : in bit; -- Adress strobe signal indicates valid adress.
            RWn                 : in bit; -- Read write control.

            DATA_IN             : in std_logic_vector (15 downto 0);
            DATA_OUT            : out std_logic_vector (15 downto 0);
            DATA_EN             : out bit;
            DTACKn              : out bit;
            
            FIFO_ERROR          : in bit;
            ACSI_DATA_REQ       : in bit;
            SECTOR_CNT_EN       : in bit;

            CD_IN               : in std_logic_vector (7 downto 0);
            CD_OUT              : out std_logic_vector (7 downto 0);
            CD_EN               : out bit;      

            CTRL_SRC_SEL        : out bit_vector(1 downto 0);
            DMA_SRC_SEL         : out bit_vector(1 downto 0);
            DMA_EN              : out bit;

            DMA_RWn             : out bit;
            HDCSn               : out bit;
            SCSICSn             : out bit;
            SDCSn               : out bit;
            FDCSn               : out bit;
            CA                  : out bit_vector(2 downto 0);
            CTRL_ACC            : out bit;

            MDET                : in bit_vector(1 downto 0);
            DISKCHNG            : in bit;
            MODE                : out bit_vector(1 downto 0);
            FCCLK               : out bit;
            
            DMA_FRAME_CNT_EN    : in bit;
            DMA_ADR             : out bit_vector(31 downto 1)
        );
    end component;

    component FDMA_SOUND_REGs
        port(
            CLK                 : in bit;
            RESET               : in bit;

            FC                  : in bit_vector(2 downto 0); -- Processor function codes.
            ADR_IN              : in bit_vector(31 downto 1); --Adress inputs.
            LDSn                : in bit; -- Lower data strobe; not used so far.
            UDSn                : in bit; -- Upper data strobe.
            ASn                 : in bit; -- Adress strobe signal indicates valid adress.
            RWn                 : in bit; -- Read write control.

            DATA_IN             : in std_logic_vector (15 downto 0);
            DATA_OUT            : out std_logic_vector (15 downto 0);
            DATA_EN             : out bit;
            DTACKn              : out bit;

            PCM_REPLAY          : out bit;
            PCM_CAPTURE         : out bit;
            
            RP_FIFO_EMPTY       : in bit;
            CA_FIFO_FULL        : in bit;
            RP_FRAME_CNT_EN     : in bit;
            CA_FRAME_CNT_EN     : in bit;
            RP_DMA_ADR          : out bit_vector(31 downto 1);
            CA_DMA_ADR          : out bit_vector(31 downto 1);
            
            -- Configuration signals:
            CROSSBAR_SOURCE_OUT     : out bit_vector(15 downto 0);
            CROSSBAR_DEST_OUT       : out bit_vector(15 downto 0);
            DAC_TRACK_SEL           : out bit_vector(1 downto 0);
            TRACK_PLAY              : out bit_vector(1 downto 0);
            SMODE_FREQ              : out bit_vector(1 downto 0);
            SMODE_SEL               : out bit_vector(1 downto 0);
            FREQ_DIV_EXT_OUT        : out bit_vector(3 downto 0);
            FREQ_DIV_INT_OUT        : out bit_vector(3 downto 0);
            REC_TRACK_SEL_OUT       : out bit_vector(1 downto 0);
            DAC_SRC                 : out bit_vector(1 downto 0);
            ADC_SRC                 : out bit_vector(1 downto 0);
            GAIN                    : out bit_vector(7 downto 0);
            ATTENUATION             : out bit_vector(7 downto 0);
            DAC_OV                  : in bit; -- Overflow.
            ADC_OV                  : in bit; -- Overflow.       
            CODEC_TAG               : out bit_vector(15 downto 0);
            CODEC_ADDRESS           : out bit_vector(15 downto 0);
            CODEC_COMMAND           : out bit_vector(15 downto 0);
            CODEC_FMODEn            : out bit; -- '0' = Falcon compatible.

            -- Interrupts:
            SCNT                : out bit; -- Timer A interrupt of the multi function port (MFP).
            SINT                : out bit -- IO7 interrupt of the multi function port (MFP).
        );
    end component;

    component PCM_CTRL
        port(
            RESET                   : in bit;

            CLK_32                   : in bit; -- 32MHz.
            SMCLK                   : in bit; -- 25.175MHz.
            CLK_EXT                 : in bit; -- External clock.

            -- External serial output channel:
            PLYDATA                 : out bit;
            PLYCLK                  : out bit;
            PLYSYNC_IN              : in bit; -- Gated clock mode.
            PLYSYNC_OUT             : out bit; -- Continuous clock mode.
            PLYSYNC_EN              : out bit; -- Continuous clock mode.

            -- External serial input channel:
            RECDATA                 : in bit;
            RECCLK                  : out bit;
            RECSYNC_IN              : in bit; -- Gated clock mode.
            RECSYNC_OUT             : out bit; -- Continuous clock mode.
            RECSYNC_EN              : out bit; -- Continuous clock mode.

            -- DSP connector:
            DSP_SRD                 : out bit; -- DSP receives data.
            DSP_SCK                 : out bit; -- Transmit clockout.
            DSP_SC2_IN              : in bit; -- Transmit syncout.
            DSP_SC2_OUT             : out bit; -- Transmit syncout.
            DSP_PLY_EN              : out bit; -- Tristate control.
            DSP_STD                 : in bit; -- DSP transmits data.
            DSP_SC0                 : out bit; -- Receive clockout.
            DSP_SC1_IN              : in bit; -- Receive syncout.
            DSP_SC1_OUT             : out bit; -- Receive syncout.
            DSP_REC_EN              : out bit; -- Tristate control.

            -- Falcon audio codec:
            SCLOCK                  : out bit;
            ASCLK                   : out bit;
            ASSYNC                  : out bit;
            ASDOUT                  : in bit; -- Sampled on the rising edge of ASCLK.
            ASDIN                   : out bit; -- Sampled on the falling edge of ASCLK.
            CODEC_ID                : out bit_vector(1 downto 0);
            
            -- PCM playback data channel:
            REPLAY_DATAREQ          : out bit; -- Data request from the sound device.
            REPLAY_DATACK           : in bit; -- Data request from the sound device.
            REPLAY_FIFO_EMPTY       : in bit;
            PCM_DATA_IN             : in bit_vector (15 downto 0);

            -- PCM capture data channel:
            CAPTURE_DATAREQ         : out bit; -- Data request from the sound device.
            CAPTURE_DATACK          : in bit; -- Data request from the sound device.
            CAPTURE_FIFO_FULL       : in bit;
            PCM_DATA_OUT            : out bit_vector (15 downto 0);

            -- Configuration signals:
            CROSSBAR_SOURCE         : in bit_vector(15 downto 0);
            CROSSBAR_DEST           : in bit_vector(15 downto 0);
            DAC_TRACK_SEL           : in bit_vector(1 downto 0);
            TRACK_PLAY              : in bit_vector(1 downto 0);
            SMODE_FREQ              : in bit_vector(1 downto 0);
            SMODE_SEL               : in bit_vector(1 downto 0);
            FREQ_DIV_EXT            : in bit_vector(3 downto 0);
            FREQ_DIV_INT            : in bit_vector(3 downto 0);
            REC_TRACK_SEL           : in bit_vector(1 downto 0);
            DAC_SRC                 : in bit_vector(1 downto 0);
            ADC_SRC                 : in bit_vector(1 downto 0);
            GAIN                    : in bit_vector(7 downto 0);
            ATTENUATION             : in bit_vector(7 downto 0);
            CODEC_TAG               : in bit_vector(15 downto 0);
            CODEC_ADDRESS           : in bit_vector(15 downto 0);
            CODEC_COMMAND           : in bit_vector(15 downto 0);
            CODEC_FMODEn            : in bit; -- '0' = Falcon compatible.
            DAC_OV                  : out bit; -- Overflow.
            ADC_OV                  : out bit -- Overflow.       
        );
    end component;

    component FDMA_MICROWIRE
        port(
            RESET           : in bit;
            CLK             : in bit; -- Use a 32MHz clock.

            FC              : in bit_vector(2 downto 0); -- Processor function codes.
            ADR             : in bit_vector(31 downto 1); --Adress inputs.
            LDSn            : in bit; -- Lower data strobe; not used so far.
            UDSn            : in bit; -- Upper data strobe.
            ASn             : in bit; -- Adress strobe signal indicates valid adress.
            RWn             : in bit; -- Read write control.

            DATA_IN         : in std_logic_vector(15 downto 0); -- Data.
            DATA_OUT        : out std_logic_vector(15 downto 0);
            DATA_EN         : out bit;
            DTACKn          : out bit;

            GPx_IN          : in bit_vector(2 downto 0);
            GPx_OUT         : out bit_vector(2 downto 0);
            GPx_EN          : out bit_vector(2 downto 0);

            MWK             : out bit;      -- Microwire clock (1MHz), GPX_OUT(2).
            MWD             : out bit;      -- Microwire data, GPX_OUT(1).
            MWEn            : buffer bit    -- Microwire enable (low active), GPX_OUT(0).
        );
    end component;
end FDMA_PKG;
