------------------------------------------------------------------------
----                                                                ----
----  Atari Falcon compatible direct memory access coprocessor.     ----
----  This file is part of the SUSKA-ATARI clone project.           ----
----                                                                ----
----  Author: Wolfgang Foerster                                     ----
----          support@inventronk.de                                 ----
----          www.inventronik.de                                    ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright (C) 2012 Inventronik GmbH                            ----
----                    Finkenstrasse 48                            ----
----                    70199 Stuttgart                             ----
----                    wf@inventronk.de                            ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Description:                                                   ----
----                                                                ----
------------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K12A  20120620 WF
--   Initial Release.
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.FDMA_PKG.all;

entity FDMA_TOP_SOC is
    generic(
        FIFO_WIDTH          : integer := 16; -- Number of bits per register.
        ACSI_FIFO_DEPTH     : integer := 32; -- Number of registers.
        REPLAY_FIFO_DEPTH   : integer := 32; -- Number of registers.
        CAPTURE_FIFO_DEPTH  : integer := 32 -- Number of registers.
    );
    port(
        RESET               : in bit;

        -- Clock system:
        CLK_32              : in bit; -- 32MHz.
        --CLK_32O           : out bit; -- 32MHz. Not used nor modeled.
        --CLK_8             : out bit; -- 8MHz. Not used nor modeled.
        --CLK_2             : out bit; -- 2MHz. Not used nor modeled.
        --BCLK              : out bit; -- 16MHz. Not used nor modeled.
        SMCLK               : in bit; -- 25.175MHz.
        CLK_EXT             : in bit; -- External clock.

        -- Misc:
        BMODE               : in bit; -- BMODE DMA control: '0' = 68030 bus timing '1' = 68000 bus timing.

        -- Adress and data bus:
        FC_IN               : in bit_vector(2 downto 0);
        FC_OUT              : out bit_vector(2 downto 0);
        ADR_IN              : in bit_vector(31 downto 1);
        ADR_OUT             : out bit_vector(31 downto 1);
        ADR_EN              : out bit;
        DATA_IN             : in std_logic_vector(15 downto 0);
        DATA_OUT            : out std_logic_vector(15 downto 0);
        DATA_EN             : out bit;

        -- Bus control signals:
        AS_INn              : in bit;
        AS_OUTn             : out bit;
        LDS_INn             : in bit;
        LDS_OUTn            : out bit;
        UDS_INn             : in bit;
        UDS_OUTn            : out bit;
        RWn_IN              : in bit;
        RWn_OUT             : out bit;
        DTACK_INn           : in bit;
        DTACK_OUTn          : out bit;

        -- Bus arbitration signals:
        BRn                 : out bit;
        BGIn                : in bit;
        BGOn                : out bit;
        BGA_INn             : in bit;
        BGA_OUTn            : out bit;
        BERRn               : in bit;

        -- ACSI bus:
        CA                  : out bit_vector(2 downto 0); -- ACSI adress.
        CR_Wn               : out bit; -- ACSI read write control.
        CR_W                : out bit; -- ACSI read write control.
        CD_IN               : in std_logic_vector(7 downto 0); -- ACSI data.
        CD_OUT              : out std_logic_vector(7 downto 0); -- ACSI data.
        CD_EN               : out bit;  -- CD data enable.

        DRIVE_SEL           : out bit_vector(1 downto 0); -- Drive selection during register access and DMA.
        FDCSn               : out bit; -- FLOPPY select.
        HDCSn               : out bit; -- ACSI drive select.
        SCSICSn             : out bit; -- SCSI drive select.
        SDCSn               : out bit; -- SDcard drive select.
        FDRQ                : in bit; -- FLOPPY request.
        HDRQ                : in bit;  -- ACSI drive request.
        ACKn                : out bit; -- ACSI data acknowledge.

        -- Floppy disk drive configuration:
        MDET                : in bit_vector(1 downto 0);
        DISKCHNG            : in bit;
        MODE                : out bit_vector(1 downto 0);
        FCCLK               : out bit; -- Floppy controller clock.

        -- External serial output channel:
        PLYDATA             : out bit;
        PLYCLK              : out bit;
        PLYSYNC_IN          : in bit; -- Gated clock mode.
        PLYSYNC_OUT         : out bit; -- Continuous clock mode.
        PLYSYNC_EN          : out bit; -- Continuous clock mode.

        -- External serial input channel:
        RECDATA             : in bit;
        RECCLK              : out bit;
        RECSYNC_IN          : in bit; -- Gated clock mode.
        RECSYNC_OUT         : out bit; -- Continuous clock mode.
        RECSYNC_EN          : out bit; -- Continuous clock mode.

        -- DSP connector:
        DSP_SRD             : out bit; -- DSP receives data.
        DSP_SCK             : out bit; -- Transmit clockout.
        DSP_SC2_IN          : in bit; -- Transmit syncout.
        DSP_SC2_OUT         : out bit; -- Transmit syncout.
        DSP_PLY_EN          : out bit; -- Tristate control.
        DSP_STD             : in bit; -- DSP transmits data.
        DSP_SC0             : out bit; -- Receive clockout.
        DSP_SC1_IN          : in bit; -- Receive syncout.
        DSP_SC1_OUT         : out bit; -- Receive syncout.
        DSP_REC_EN          : out bit; -- Tristate control.

        -- Falcon audio codec:
        SCLOCK              : out bit;
        ASCLK               : out bit;
        ASSYNC              : out bit;
        ASDOUT              : in bit; -- Sampled on the rising edge of ASCLK.
        ASDIN               : out bit; -- Sampled on the falling edge of ASCLK.
        CODEC_ID            : out bit_vector(1 downto 0);

        -- General Purpose I/O:
        GPIO_IN             : in bit_vector(2 downto 0);
        GPIO_OUT            : out bit_vector(2 downto 0);
        GPIO_EN             : out bit_vector(2 downto 0);
        
        -- Interrupt signals:
        SCNT                : out bit; -- Timer A interrupt of the multi function port (MFP).
        SINT                : out bit; -- IO7 interrupt of the multi function port (MFP).
        HDINTn              : in bit;
        FDINT               : in bit;
        DSKIRQn             : out bit; -- Open collector with weak pull up?
        
        -- Microwire Interface:
        MWK                 : out bit;
        MWD                 : out bit;
        MWEn                : out bit
    );
end entity FDMA_TOP_SOC;


architecture structure of FDMA_TOP_SOC is
signal ADR_OUT_ACSI         : bit_vector(31 downto 1);
signal ADR_EN_ACSI          : bit;
signal ADR_OUT_REPLAY       : bit_vector(31 downto 1);
signal ADR_EN_REPLAY        : bit;
signal ADR_OUT_CAPTURE      : bit_vector(31 downto 1);
signal ADR_EN_CAPTURE       : bit;
signal DATA_OUT_MW          : std_logic_vector(15 downto 0);
signal DATA_EN_MW           : bit;
signal DATA_OUT_ACSI_REG    : std_logic_vector(15 downto 0);
signal DATA_EN_ACSI_REG     : bit;
signal DATA_OUT_SOUND_REG   : std_logic_vector(15 downto 0);
signal DATA_EN_SOUND_REG    : bit;
signal DATA_OUT_FIFOs       : std_logic_vector(15 downto 0);
signal DATA_EN_ACSI         : bit;
signal DATA_EN_CAPTURE      : bit;
signal CA_I                 : bit_vector(2 downto 0);
signal CD_OUT_REG           : std_logic_vector(7 downto 0);
signal CD_EN_REG            : bit;
signal CD_OUT_FIFO          : std_logic_vector(7 downto 0);
signal DTACK_MWn            : bit;
signal DTACK_ACSI_REGn      : bit;
signal DTACK_SOUND_REGn     : bit;
signal CODEC_INT            : bit;
signal CODEC_LOAD           : bit;
signal SDATA_L              : bit_vector(19 downto 0);
signal SDATA_R              : bit_vector(19 downto 0);
signal CD_RD_HIn            : bit;
signal CD_RD_LOWn           : bit;
signal CD_HIBUF_EN          : bit;
signal ACSI_FIFO_CLRn       : bit;
signal ACSI_FIFO_WR         : bit;
signal ACSI_FIFO_RD         : bit;
signal ACSI_FIFO_FULL       : bit;
signal ACSI_FIFO_LOW        : bit;
signal ACSI_FIFO_EMPTY      : bit;
signal REPLAY_FIFO_CLRn     : bit;
signal REPLAY_FIFO_WR       : bit;
signal REPLAY_FIFO_RD       : bit;
signal REPLAY_FIFO_FULL     : bit;
signal REPLAY_FIFO_LOW      : bit;
signal REPLAY_FIFO_EMPTY    : bit;
signal CAPTURE_FIFO_CLRn    : bit;
signal CAPTURE_FIFO_WR      : bit;
signal CAPTURE_FIFO_RD      : bit;
signal CAPTURE_FIFO_FULL    : bit;
signal CAPTURE_FIFO_LOW     : bit;
signal CAPTURE_FIFO_EMPTY   : bit;
signal DMA_RWn              : bit;
signal SECTOR_CNT_EN        : bit;
signal ACSI_DATA_REQ        : bit;
signal DMA_FRAME_CNT_EN     : bit;
signal CTRL_ACC             : bit;
signal FDCS_CTRL_ACCn       : bit;
signal FDCS_DMA_ACCn        : bit;
signal DMA_SRC_SEL          : bit_vector(1 downto 0);
signal CTRL_SRC_SEL         : bit_vector(1 downto 0);
signal DMA_EN               : bit;
signal ACSI_FIFO_ERROR      : bit;
signal PCM_REPLAY           : bit;
signal PCM_CAPTURE          : bit;
signal REPLAY_FRAME_CNT_EN  : bit;
signal CAPTURE_FRAME_CNT_EN : bit;
signal REPLAY_DATAREQ       : bit;
signal REPLAY_DATACK        : bit;
signal REPLAY_DATA_OUT      : bit_vector(15 downto 0);
signal CAPTURE_DATAREQ      : bit;
signal CAPTURE_DATACK       : bit;
signal CAPTURE_DATA_IN      : bit_vector(15 downto 0);
signal CROSSBAR_SOURCE      : bit_vector(15 downto 0);
signal CROSSBAR_DEST        : bit_vector(15 downto 0);
signal DAC_TRACK_SEL        : bit_vector(1 downto 0);
signal TRACK_PLAY           : bit_vector(1 downto 0);
signal SMODE_SEL            : bit_vector(1 downto 0);
signal SMODE_FREQ           : bit_vector(1 downto 0);
signal FREQ_DIV_EXT         : bit_vector(3 downto 0);
signal FREQ_DIV_INT         : bit_vector(3 downto 0);
signal REC_TRACK_SEL        : bit_vector(1 downto 0);
signal DAC_SRC              : bit_vector(1 downto 0);
signal ADC_SRC              : bit_vector(1 downto 0);
signal GAIN                 : bit_vector(7 downto 0);
signal ATTENUATION          : bit_vector(7 downto 0);
signal DAC_OV               : bit;
signal ADC_OV               : bit;
signal CODEC_TAG            : bit_vector(15 downto 0);
signal CODEC_ADDRESS        : bit_vector(15 downto 0);
signal CODEC_COMMAND        : bit_vector(15 downto 0);
signal CODEC_FMODEn         : bit;
begin
    ADR_EN <= ADR_EN_ACSI or ADR_EN_REPLAY or ADR_EN_CAPTURE;
    ADR_OUT <= ADR_OUT_ACSI when ADR_EN_ACSI = '1' else
               ADR_OUT_REPLAY when ADR_EN_REPLAY = '1' else ADR_OUT_CAPTURE;
               
    DATA_EN <= DATA_EN_ACSI_REG or DATA_EN_SOUND_REG or DATA_EN_MW or DATA_EN_ACSI or DATA_EN_CAPTURE;
    DATA_OUT <= DATA_OUT_ACSI_REG when DATA_EN_ACSI_REG = '1' else
                DATA_OUT_SOUND_REG when DATA_EN_SOUND_REG = '1' else    
                DATA_OUT_MW when DATA_EN_MW = '1' else
                DATA_OUT_FIFOs when DATA_EN_ACSI = '1' else
                DATA_OUT_FIFOs when DATA_EN_CAPTURE = '1' else (others => '0');

    DTACK_OUTn <= DTACK_ACSI_REGn and DTACK_SOUND_REGn and DTACK_MWn;

    FDCSn <= FDCS_CTRL_ACCn and FDCS_DMA_ACCn;

    CR_Wn <= RWn_IN when CTRL_ACC = '1' else DMA_RWn;
    CR_W <= not RWn_IN when CTRL_ACC = '1' else not DMA_RWn;

    CA(2) <= CA_I(2) when CTRL_ACC = '1' else '1'; -- Default is required for DMA operation.
    CA(1) <= CA_I(1) when CTRL_ACC = '1' else '1';
    CA(0) <= CA_I(0);

    CD_OUT <= CD_OUT_REG when CD_EN_REG = '1' else CD_OUT_FIFO;
    CD_EN <= '1' when CD_EN_REG = '1' else
             '1' when CD_RD_HIn = '0' or CD_RD_LOWn = '0' else '0';

    -- Decoding for DRIVE_SEL:
    -- ACSI = "00", SCSI = "01", Floppy = "10", SD card = "11".
    DRIVE_SEL <= CTRL_SRC_SEL when CTRL_ACC = '1' else DMA_SRC_SEL;

    DSKIRQn <= '1' when HDINTn = '1' and FDINT = '0' else '0';

    DMA_CONTROL: FDMA_CTRL
        port map(
            CLK                 => CLK_32,
            RESET               => RESET,

            -- Bus control:
            BMODE               => BMODE,
            FC_OUT              => FC_OUT,
            AS_INn              => AS_INn,
            AS_OUTn             => AS_OUTn,
            ADR_EN_ACSI         => ADR_EN_ACSI,
            ADR_EN_REPLAY       => ADR_EN_REPLAY,
            ADR_EN_CAPTURE      => ADR_EN_CAPTURE,
            UDSn                => UDS_OUTn,
            LDSn                => LDS_OUTn,
            RWn                 => RWn_OUT,
            BERRn               => BERRn,
            DTACKn              => DTACK_INn,
            DATA_EN_ACSI        => DATA_EN_ACSI,
            DATA_EN_CAPTURE     => DATA_EN_CAPTURE,

            -- Bus arbitration:
            BRn                 => BRn,
            BGIn                => BGIn,
            BGOn                => BGOn,
            BGACK_INn           => BGA_INn,
            BGACK_OUTn          => BGA_OUTn,

            -- Control signals for ACSI:
            ACSI_FIFO_CLRn      => ACSI_FIFO_CLRn,
            ACSI_FIFO_WR        => ACSI_FIFO_WR,
            ACSI_FIFO_RD        => ACSI_FIFO_RD,
            ACSI_FIFO_FULL      => ACSI_FIFO_FULL,
            ACSI_FIFO_LOW       => ACSI_FIFO_LOW,
            ACSI_FIFO_EMPTY     => ACSI_FIFO_EMPTY,

            -- Control signals for the PCM out:
            REPLAY_FIFO_CLRn    => REPLAY_FIFO_CLRn,
            REPLAY_FIFO_WR      => REPLAY_FIFO_WR,
            REPLAY_FIFO_RD      => REPLAY_FIFO_RD,
            REPLAY_FIFO_FULL    => REPLAY_FIFO_FULL,
            REPLAY_FIFO_LOW     => REPLAY_FIFO_LOW,

            -- Control signals for the PCM in:
            CAPTURE_FIFO_CLRn   => CAPTURE_FIFO_CLRn,
            CAPTURE_FIFO_WR     => CAPTURE_FIFO_WR,
            CAPTURE_FIFO_RD     => CAPTURE_FIFO_RD,
            CAPTURE_FIFO_LOW    => CAPTURE_FIFO_LOW,
            CAPTURE_FIFO_EMPTY  => CAPTURE_FIFO_EMPTY,

            -- Handshaking:
            HDRQ                => HDRQ,
            HD_ACKn             => ACKn,
            FDCRQ               => FDRQ,
            FDCS_DMA_ACCn       => FDCS_DMA_ACCn,
            ACSI_DATA_REQ       => ACSI_DATA_REQ,
            REPLAY_DATAREQ      => REPLAY_DATAREQ,
            REPLAY_DATACK       => REPLAY_DATACK,
            CAPTURE_DATAREQ     => CAPTURE_DATAREQ,
            CAPTURE_DATACK      => CAPTURE_DATACK,

            SECTOR_CNT_EN       => SECTOR_CNT_EN,
            DMA_FRAME_CNT_EN    => DMA_FRAME_CNT_EN,
            RP_FRAME_CNT_EN     => REPLAY_FRAME_CNT_EN,
            CA_FRAME_CNT_EN     => CAPTURE_FRAME_CNT_EN,

            -- Control signals for the ACSI multiplexer:
            CD_HIBUF_EN         => CD_HIBUF_EN,
            CD_RD_HIn           => CD_RD_HIn,
            CD_RD_LOWn          => CD_RD_LOWn,

            -- Other controls:
            DMA_EN              => DMA_EN,
            PCM_REPLAY          => PCM_REPLAY,
            PCM_CAPTURE         => PCM_CAPTURE,
            
            DMA_RWn             => DMA_RWn,
            DMA_SRC_SEL         => DMA_SRC_SEL
        );

    I_FIFOs: DMA_FIFOs
        generic map(
            FIFO_WIDTH          => FIFO_WIDTH,
            ACSI_FIFO_DEPTH     => ACSI_FIFO_DEPTH,
            REPLAY_FIFO_DEPTH   => REPLAY_FIFO_DEPTH,
            CAPTURE_FIFO_DEPTH  => CAPTURE_FIFO_DEPTH
            )
        port map(
            RESET               => RESET,
            CLK                 => CLK_32,

            DATA_IN             => DATA_IN,
            DATA_OUT            => DATA_OUT_FIFOs,

            CD_DATA_IN          => CD_IN,
            CD_DATA_OUT         => CD_OUT_FIFO,

            CAPTURE_DATA_IN     => CAPTURE_DATA_IN,
            REPLAY_DATA_OUT     => REPLAY_DATA_OUT,

            DMA_RWn             => DMA_RWn,
            
            -- Control signals for the ACSI multiplexer:
            ACSI_DATA_EN        => DATA_EN_ACSI,
            CD_HIBUF_EN         => CD_HIBUF_EN,
            CD_RD_HIn           => CD_RD_HIn,
            CD_RD_LOWn          => CD_RD_LOWn,
            CAPTURE_DATA_EN     => DATA_EN_CAPTURE,

            -- Control signals for the ACSI bus:
            ACSI_FIFO_CLRn      => ACSI_FIFO_CLRn,
            ACSI_FIFO_WR        => ACSI_FIFO_WR,
            ACSI_FIFO_RD        => ACSI_FIFO_RD,
            ACSI_FIFO_FULL      => ACSI_FIFO_FULL,
            ACSI_FIFO_LOW       => ACSI_FIFO_LOW,
            ACSI_FIFO_EMPTY     => ACSI_FIFO_EMPTY,
            ACSI_FIFO_ERROR     => ACSI_FIFO_ERROR,

            -- Control signals for the PCM out:
            REPLAY_FIFO_CLRn    => REPLAY_FIFO_CLRn,
            REPLAY_FIFO_WR      => REPLAY_FIFO_WR,
            REPLAY_FIFO_RD      => REPLAY_FIFO_RD,
            REPLAY_FIFO_FULL    => REPLAY_FIFO_FULL,
            REPLAY_FIFO_LOW     => REPLAY_FIFO_LOW,
            REPLAY_FIFO_EMPTY   => REPLAY_FIFO_EMPTY,

            -- Control signals for the PCM in:
            CAPTURE_FIFO_CLRn   => CAPTURE_FIFO_CLRn,
            CAPTURE_FIFO_WR     => CAPTURE_FIFO_WR,
            CAPTURE_FIFO_RD     => CAPTURE_FIFO_RD,
            CAPTURE_FIFO_FULL   => CAPTURE_FIFO_FULL,
            CAPTURE_FIFO_LOW    => CAPTURE_FIFO_LOW,
            CAPTURE_FIFO_EMPTY  => CAPTURE_FIFO_EMPTY
        );

    ACSI_DMA_REGS: FDMA_ACSI_REGs
        port map(
            CLK                 => CLK_32,
            RESET               => RESET,

            FC                  => FC_IN,
            ADR_IN              => ADR_IN,
            LDSn                => LDS_INn,
            UDSn                => UDS_INn,
            ASn                 => AS_INn,
            RWn                 => RWn_IN,

            DATA_IN             => DATA_IN,
            DATA_OUT            => DATA_OUT_ACSI_REG,
            DATA_EN             => DATA_EN_ACSI_REG,
            DTACKn              => DTACK_ACSI_REGn,
            
            FIFO_ERROR          => ACSI_FIFO_ERROR,
            ACSI_DATA_REQ       => ACSI_DATA_REQ,
            SECTOR_CNT_EN       => SECTOR_CNT_EN,

            CD_IN               => CD_IN,
            CD_OUT              => CD_OUT_REG,
            CD_EN               => CD_EN_REG,

            CTRL_SRC_SEL        => CTRL_SRC_SEL,
            DMA_SRC_SEL         => DMA_SRC_SEL,
            DMA_EN              => DMA_EN,

            DMA_RWn             => DMA_RWn,
            HDCSn               => HDCSn,
            SCSICSn             => SCSICSn,
            SDCSn               => SDCsn,
            FDCSn               => FDCS_CTRL_ACCn,
            CA                  => CA_I,
            CTRL_ACC            => CTRL_ACC,

            MODE                => MODE,
            MDET                => MDET,
            DISKCHNG            => DISKCHNG,
            FCCLK               => FCCLK,
            
            DMA_FRAME_CNT_EN    => DMA_FRAME_CNT_EN,
            DMA_ADR             => ADR_OUT_ACSI
        );

    SOUND_DMA_REGs: FDMA_SOUND_REGs
        port map(
            CLK                 => CLK_32,
            RESET               => RESET,

            FC                  => FC_IN,
            ADR_IN              => ADR_IN,
            LDSn                => LDS_INn,
            UDSn                => UDS_INn,
            ASn                 => AS_INn,
            RWn                 => RWn_IN,

            DATA_IN             => DATA_IN,
            DATA_OUT            => DATA_OUT_SOUND_REG,
            DATA_EN             => DATA_EN_SOUND_REG,
            DTACKn              => DTACK_SOUND_REGn,
            
            PCM_REPLAY          => PCM_REPLAY,
            PCM_CAPTURE         => PCM_CAPTURE,

            RP_FIFO_EMPTY       => REPLAY_FIFO_EMPTY,
            CA_FIFO_FULL        => CAPTURE_FIFO_FULL,
            RP_FRAME_CNT_EN     => REPLAY_FRAME_CNT_EN,
            CA_FRAME_CNT_EN     => CAPTURE_FRAME_CNT_EN,
            RP_DMA_ADR          => ADR_OUT_REPLAY,
            CA_DMA_ADR          => ADR_OUT_CAPTURE,
            
            CROSSBAR_SOURCE_OUT => CROSSBAR_SOURCE,
            CROSSBAR_DEST_OUT   => CROSSBAR_DEST,
            DAC_TRACK_SEL       => DAC_TRACK_SEL,
            TRACK_PLAY          => TRACK_PLAY,
            SMODE_FREQ          => SMODE_FREQ,
            SMODE_SEL           => SMODE_SEL,
            FREQ_DIV_EXT_OUT    => FREQ_DIV_EXT,
            FREQ_DIV_INT_OUT    => FREQ_DIV_INT,
            REC_TRACK_SEL_OUT   => REC_TRACK_SEL,
            DAC_SRC             => DAC_SRC,
            ADC_SRC             => ADC_SRC,
            GAIN                => GAIN,
            ATTENUATION         => ATTENUATION,
            DAC_OV              => DAC_OV,
            ADC_OV              => ADC_OV,
            CODEC_TAG           => CODEC_TAG,
            CODEC_ADDRESS       => CODEC_ADDRESS,
            CODEC_COMMAND       => CODEC_COMMAND,
            CODEC_FMODEn        => CODEC_FMODEn,

            SCNT                => SCNT,
            SINT                => SINT
        );

    I_PCM_CTRL: PCM_CTRL
        port map(
            RESET               => RESET,

            CLK_32              => CLK_32,
            SMCLK               => SMCLK,
            CLK_EXT             => CLK_EXT,

            PLYDATA             => PLYDATA,
            PLYCLK              => PLYCLK,
            PLYSYNC_IN          => PLYSYNC_IN,
            PLYSYNC_OUT         => PLYSYNC_OUT,
            PLYSYNC_EN          => PLYSYNC_EN,

            RECDATA             => RECDATA,
            RECCLK              => RECCLK,
            RECSYNC_IN          => RECSYNC_IN,
            RECSYNC_OUT         => RECSYNC_OUT,
            RECSYNC_EN          => RECSYNC_EN,

            DSP_SRD             => DSP_SRD,
            DSP_SCK             => DSP_SCK,
            DSP_SC2_IN          => DSP_SC2_IN,
            DSP_SC2_OUT         => DSP_SC2_OUT,
            DSP_PLY_EN          => DSP_PLY_EN,
            DSP_STD             => DSP_STD,
            DSP_SC0             => DSP_SC0,
            DSP_SC1_IN          => DSP_SC1_IN,
            DSP_SC1_OUT         => DSP_SC1_OUT,
            DSP_REC_EN          => DSP_REC_EN,

            SCLOCK              => SCLOCK,
            ASCLK               => ASCLK,
            ASSYNC              => ASSYNC,
            ASDOUT              => ASDOUT,
            ASDIN               => ASDIN,
            CODEC_ID            => CODEC_ID,

            REPLAY_DATAREQ      => REPLAY_DATAREQ,
            REPLAY_DATACK       => REPLAY_DATACK,
            REPLAY_FIFO_EMPTY   => REPLAY_FIFO_EMPTY,
            PCM_DATA_IN         => REPLAY_DATA_OUT,
            CAPTURE_DATAREQ     => CAPTURE_DATAREQ,
            CAPTURE_DATACK      => CAPTURE_DATACK,
            CAPTURE_FIFO_FULL   => CAPTURE_FIFO_FULL,
            PCM_DATA_OUT        => CAPTURE_DATA_IN,
                               
            CROSSBAR_SOURCE     => CROSSBAR_SOURCE,
            CROSSBAR_DEST       => CROSSBAR_DEST,
            DAC_TRACK_SEL       => DAC_TRACK_SEL,
            TRACK_PLAY          => TRACK_PLAY,
            SMODE_FREQ          => SMODE_FREQ,
            SMODE_SEL           => SMODE_SEL,
            FREQ_DIV_EXT        => FREQ_DIV_EXT,
            FREQ_DIV_INT        => FREQ_DIV_INT,
            REC_TRACK_SEL       => REC_TRACK_SEL,
            DAC_SRC             => DAC_SRC,
            ADC_SRC             => ADC_SRC,
            GAIN                => GAIN,
            ATTENUATION         => ATTENUATION,
            CODEC_TAG           => CODEC_TAG,
            CODEC_ADDRESS       => CODEC_ADDRESS,
            CODEC_COMMAND       => CODEC_COMMAND,
            CODEC_FMODEn        => CODEC_FMODEn,
            DAC_OV              => DAC_OV,
            ADC_OV              => ADC_OV           
        );

    I_MICROWIRE: FDMA_MICROWIRE
        port map(
            RESET               => RESET,
            CLK                 => CLK_32,

            FC                  => FC_IN,
            ADR                 => ADR_IN,
            LDSn                => LDS_INn,
            UDSn                => UDS_INn,
            ASn                 => AS_INn,
            RWn                 => RWn_IN,

            DATA_IN             => DATA_IN,
            DATA_OUT            => DATA_OUT_MW,
            DATA_EN             => DATA_EN_MW,
            DTACKn              => DTACK_MWn,

            GPx_IN              => GPIO_IN,
            GPx_OUT             => GPIO_OUT,
            GPx_EN              => GPIO_EN,

            MWK                 => MWK,
            MWD                 => MWD,
            MWEn                => MWEn
        );
    
end Structure;