----------------------------------------------------------------------
----                                                              ----
---- Atari STE compatible IP Core                                  ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- This model provides the top level file of an STE compatible  ----
---- machine including CPU, Blitter, Shadow, MCU, DMA,            ----
---- Shifter, GLUE, MFP, SOUND, ACIA and RTC.                     ----
----                                                              ----
---- Important Notice concerning the clock system:                ----
---- The systems of the original ST or STE machines used several  ----
---- clocks which must stand in a fixed relation to each other.   ----
---- This core uses one central system clock of 16MHz. From this  ----
---- clock all required clocks are derived. These are CLK_1,      ----
---- CLK_2 and CLK_CPU. These are the clocks used for clocking    ----
---- D type flip-flops of the system and are provided by a first  ----
---- phase locked loop circuit.                                   ----
---- The clocks should used as follows:                           ----
---- CLK_1 for the CPU, BLITTER, GLUE, DMA, MFP, 1772,            ----
---- SOUND and the UARTs.                                         ----
---- CLK_2 for the MCU, , SHIFTER, the video RAM component and    ----
---- the external SD-RAMs and the memory data buffer.             ----
---- CLK_CPU for the processor unit.                              ----
---- Beside these 'real' clocks, there are several auxiliary      ----
---- clocks which are processed by one of the above mentioned     ----
---- clocks (CLK_1, CLK_2 and CLK_CPU). There are two clocks      ----
---- SCLK_6M4 and CLK_24576 which are provided by a second phase  ----
---- locked loop and a counter process. The SCLK_6M4 controls the ----
---- DMA sound module whereas the CLK_24576 is responsible for a  ----
---- correct timing of the MFP. The clocks are outputs of the     ----
---- counter process for their frequency is too low to provide it ----
---- directly by a phase locked loop.                             ----
---- Last but not least there are two clocks CLK_2M0 and CLK_0M5  ----
---- which control the UARTs and the sound chip. These two clocks ----
---- are provided by a counter out of the CLK_PLL_16000 due to    ----
---- their low frequency.                                         ----
---- To meet the correct timing requirements for the following    ----
---- units, there must be the correct generic settings in the     ----
---- respective top level _soc files.                             ----
---- The timing critical units are:                               ----
----   - the video timing in the GLUE.                            ----
----   - the paddle counter in the GLUE.                          ----
----   - the video phase control in the MCU.                      ----
----   - the dma sound control in the MCU.                        ----
----                                                              ----
---- The phase locked loops and the video memory for the LCD are  ----
---- customer / hardware specific components and therefore        ----
---- declared in this top level file. This kind of modelling has  ----
---- the advantage, that the migration to other hardware is simple----
---- by modifying only the top level file of the SUSKA core.      ----
----                                                              ----
---- SD-RAM section: This core provides a memory of 64MByte. This ----
----   is achieved by SDRAM types with 4 different RAM banks. The ----
----   memory organisation is 2 main banks consisting of 1 SDRAM  ----
----   chip each. Each chip features four RAM banks with at least ----
----   4M words. So in total we have 2*4*4M words.                ----
----                                                              ----
---- Recommendations for the signal termination:                  ----
----   The following signals should be terminated with a weak     ----
----   pull up resistor (~22K). Use either weak pull up resistors ----
----   of the FPGAs or external ones:                             ----
----     DATA, ADR, FC.                                           ----
----   Use for the following signals strong pull up resistors     ----
----   with a value of about 6K8:                                 ----
----     ASn, RWn, UDSn, LDSn, DTACKn, BGACKn, BRn,               ----
----     EINT3, EINT5, EINT7.                                     ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright � 2009-2010 Wolfgang Foerster Inventronik GmbH.    ----
---- All rights reserved. No portion of this sourcecode may be    ----
---- reproduced or transmitted in any form by any means, whether  ----
---- by electronic, mechanical, photocopying, recording or        ----
---- otherwise, without my written permission.                    ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K9B  2009/12/24 WF
--   Initial Release.
--

library work;
use work.WF_SUSKA_CORE_B_STE_PKG.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WF_SUSKA_TOP is
    port(
        -- System controls:
        RESET_COREn             : in bit; -- FPGA reset.
        RESETn                  : inout std_logic; -- System and CPU reset.
        CLK_PLL1                : in std_logic; -- 16 MHz system clock.
        CLK_PLL2                : in std_logic; -- 16 MHz system clock.

        -- Data and address busses:
        DATA                    : inout std_logic_vector(15 downto 0);
        ADR                     : out std_logic_vector(19 downto 1);

        -- The SDRAM and ROM interface:
        RAM_CLK                 : out std_logic;
        RAM_CKE                 : out bit;

        RAM_RASn                : out bit;
        RAM_CASn                : out bit;
        RAM_DQMH                : out bit;
        RAM_DQML                : out bit;
        RAM_WEn                 : out bit;
        RAM_ADR                 : out bit_vector(12 downto 0);
        RAM_BA                  : out bit_vector(1 downto 0);
        RAM_DATA                : inout std_logic_vector(15 downto 0);

        -- Video interface:
        CRT_R                   : out bit_vector(7 downto 4);
        CRT_G                   : out bit_vector(7 downto 4);
        CRT_B                   : out bit_vector(7 downto 4);

        HSYNC                   : inout std_logic;
        VSYNC                   : inout std_logic;

        -- Keyboard:
        KEYB_RxD                : in bit;
        KEYB_TxD                : out std_logic;
        
        -- DS1392 RTC:
        DS1392_D                : inout std_logic;
        DS1392_SCL              : out bit;
        DS1392_CE               : out bit;

        -- Flash controls:
        FLASH_RDY               : in bit;
        FLASH_ADR22             : out bit;
        FLASH_ADR21             : out bit;
        FLASH_ADR20             : out bit;
        FLASH_WEn               : out bit;
        FLASH_OEn               : out bit;
        FLASH_CEn               : out bit;

        -- Audio DAC AD5302:
        DAC_SCLK                : out bit;
        DAC_SYNCn               : out bit;
        DAC_SDATA               : out bit;
        DAC_LDACn               : out bit;

        -- Other system signals:
        LED3                    : out bit; -- Indicates unlocked PLLs.
        LED2                    : out bit; -- SD card and Boot loader active...
        LED1                    : out bit; -- SD card LED.
        JOY_DIS                 : out bit; -- Disables the output buffer of the joystick.
        CONFIG                  : in bit_vector(6 downto 1); -- Configuration switches, use weak pull up.
        JOY_A5                  : in bit; -- Joystick pin.
        JOY_A2                  : in bit; -- Joystick pin.
        
        -- SD card and SPI interface:
        SDC_SPI_CLK             : inout std_logic;
        MOSI                    : inout std_logic; -- SD-Drive is master, bootloader is slave.
        MISO                    : inout std_logic; -- SD-Drive is master, bootloader is slave.
        SDC_D                   : inout std_logic_vector(3 downto 1); -- SD card interconnect.
        SDC_CDn                 : in bit; -- Card detect..
        SDC_WP                  : in bit; -- Write protect.
        SDC_PWRn                : out bit; -- Power switch.

        -- Microcontroller interface:
        SDC_BSYn                : out bit;
        BOOT_ACK                : in bit;
        BOOT_REQ                : out bit
    );
end entity WF_SUSKA_TOP;

architecture STRUCTURE of WF_SUSKA_TOP is
-- Hardware specific components:
component cyclone3_pll_1
    PORT
    (
        areset        : IN STD_LOGIC  := '0';
        inclk0        : IN STD_LOGIC  := '0';
        c0            : OUT STD_LOGIC ;
        c1            : OUT STD_LOGIC ;
        c2            : OUT STD_LOGIC ;
        c3            : OUT STD_LOGIC ;
        locked        : OUT STD_LOGIC 
    );
end component;

component cyclone3_pll_2
    PORT
    (
        areset        : IN STD_LOGIC  := '0';
        inclk0        : IN STD_LOGIC  := '0';
        c0            : OUT STD_LOGIC ;
        c1            : OUT STD_LOGIC ;
        c2            : OUT STD_LOGIC ;
        locked        : OUT STD_LOGIC 
    );
end component;

signal RESET_Sn                     : bit;
signal RESET_CORE_Sn                : bit;
signal RESET_CPUn                   : bit;
signal RESET_INn                    : bit;
signal RESET_EN_68K00               : bit;
signal RESET_BOOTn                  : bit;
signal RESET_MCUn                   : bit;
signal HALT_INn                     : std_logic;
signal PLL1_LOCKED                  : std_logic;
signal PLL2_LOCKED                  : std_logic;
signal CLK_PLL_IN0                  : std_logic;
signal CLK_PLL2_IN0                 : std_logic;
signal CLK_PLL_CPU                  : std_logic;
signal CLK_PLL_4x                   : std_logic;
signal CLK_PLL_2x                   : std_logic;
signal CLK_PLL_1x                   : std_logic;
signal CLK_PLL_256                  : std_logic;
signal CLK_PLL_394                  : std_logic;
signal CLK_PLL_16000                : std_logic;
signal CLK_CPU                      : bit;
signal CLK_4                        : bit;
signal CLK_2                        : bit;
signal CLK_1                        : bit;
signal CLK_2M0                      : bit;
signal CLK_0M5                      : bit;
signal CLK_24576                    : bit;
signal SCLK_6M4                     : bit;
signal DATA_OUT_68K00               : std_logic_vector(15 downto 0);
signal DATA_EN_68K00                : bit;
signal DATA_OUT_BLT                 : std_logic_vector(15 downto 0);
signal DATA_EN_BLT                  : bit;
signal DATA_OUT_GLUE                : std_logic_vector(15 downto 0);
signal DATA_EN_GLUE                 : bit;
signal DATA_OUT_MCU                 : std_logic_vector(7 downto 0);
signal DATA_EN_MCU                  : bit;
signal DATA_OUT_DMA                 : std_logic_vector(15 downto 0);
signal DATA_EN_DMA                  : bit;
signal DATA_OUT_MFP                 : std_logic_vector(7 downto 0);
signal DATA_EN_MFP                  : bit;
signal DATA_OUT_SOUND               : std_logic_vector(7 downto 0);
signal DATA_EN_SOUND                : bit;
signal DATA_OUT_ACIA_I              : std_logic_vector(7 downto 0);
signal DATA_EN_ACIA_I               : bit;
signal DATA_OUT_RP5C15              : std_logic_vector(3 downto 0);
signal DATA_EN_RP5C15               : bit;
signal DATA_OUT_BOOT                : std_logic_vector(15 downto 0);
signal DATA_EN_BOOT                 : bit;
signal ADR_IN                       : bit_vector(23 downto 1);
signal ADR_OUT                      : std_logic_vector(23 downto 1);
signal ADR_OUT_68K00                : std_logic_vector(23 downto 1);
signal ADR_EN_68K00                 : bit;
signal ADR_OUT_BLT                  : std_logic_vector(23 downto 1);
signal ADR_EN_BLT                   : bit;
signal ADR_OUT_BOOT                 : std_logic_vector(24 downto 1);
signal ADR_EN_BOOT                  : bit;
signal MCU_ADR                      : bit_vector(23 downto 1);
signal MDAT_BUFFER                  : std_logic_vector(15 downto 0);
signal BERRn                        : bit;
signal AS_INn                       : bit;
signal UDS_INn                      : bit;
signal LDS_INn                      : bit;
signal RWn_IN                       : bit;
signal RWn_OUT_68K00                : bit;
signal RW_OUT_EN_68K00              : bit;
signal AS_OUT_68K00n                : bit;
signal AS_OUT_EN_68K00              : bit;
signal UDS_OUT_68K00n               : bit;
signal UDS_OUT_EN_68K00             : bit;
signal LDS_OUT_68K00n               : bit;
signal LDS_OUT_EN_68K00             : bit;
signal RWn_OUT_BLT                  : bit;
signal AS_OUT_BLTn                  : bit;
signal UDS_OUT_BLTn                 : bit;
signal LDS_OUT_BLTn                 : bit;
signal BUSCTRL_EN_BLT               : bit;
signal RWn_OUT_GLUE                 : bit;
signal AS_OUT_GLUEn                 : bit;
signal UDS_OUT_GLUEn                : bit;
signal LDS_OUT_GLUEn                : bit;
signal BUSCTRL_EN_GLUE              : bit;
signal ROM2n                        : bit;
signal ROM3n                        : bit;
signal ROM4n                        : bit;
signal ACIA_CS_I                    : bit;
signal MFP_CS_In                    : bit;
signal SNDCS_I                      : bit;
signal SNDIR_I                      : bit;
signal FCS_In                       : bit;
signal RAM_In                       : bit;
signal DMA_In                       : bit;
signal DEV_In                       : bit;
signal FC_IN                        : bit_vector(2 downto 0);
signal FC_OUT_68K00                 : std_logic_vector(2 downto 0);
signal FC_OUT_EN_68K00              : bit;
signal FC_OUT_BLT                   : std_logic_vector(2 downto 0);
signal FC_OUT_GLUE                  : bit_vector(2 downto 0);
signal FCLK                         : bit;
signal DTACK_In                     : bit;
signal DTACK_INn                    : bit;
signal DTACK_OUT_BLTn               : bit;
signal DTACK_OUT_GLUEn              : bit;
signal DTACK_OUT_MCUn               : bit;
signal DTACK_OUT_MFPn               : bit;
signal IPLn                         : bit_vector(2 downto 0);
signal E_I                          : bit;
signal VPAn                         : bit;
signal VMAn                         : bit;
signal BR_In                        : bit;
signal BR_BLTn                      : bit;
signal BR_GLUEn                     : bit;
signal BG_68K00n                    : bit;
signal BG_BLTn                      : bit;
signal BGACK_GLUEn                  : bit;
signal BGACK_BLTn                   : bit;
signal INT_BLTn                     : bit;
signal RP5C15_CSn                   : bit;
signal RP5C15_WRn                   : bit;
signal RP5C15_RDn                   : bit;
signal HSYNC_In                     : bit;
signal HSYNC_On                     : bit;
signal VSYNC_In                     : bit;
signal VSYNC_On                     : bit;
signal SYNC_EN                      : bit;
signal BLANKn                       : bit;
signal MONOCHROME                   : bit;
signal AVECn                        : bit;
signal MFPINTn                      : bit;
signal IACKn                        : bit;
signal RDY_DMAn                     : bit;
signal RDY_GLUEn                    : bit;
signal DINTn                        : bit;
signal DCYCn                        : bit;
signal CMPCSn                       : bit;
signal WDATn                        : bit;
signal RDATn                        : bit;
signal LATCHn                       : bit;
signal CA1                          : bit;
signal CA2                          : bit;
signal CR_Wn                        : bit;
signal CD_IN                        : std_logic_vector(7 downto 0);
signal CD_OUT_DMA                   : std_logic_vector(7 downto 0);
signal CD_EN_DMA                    : bit;
signal CD_OUT_FDC                   : std_logic_vector(7 downto 0);
signal CD_OUT_DRIVES                : std_logic_vector(7 downto 0);
signal CD_EN_DRIVES                 : bit;
signal CD_EN_FDC                    : bit;
signal SINT_IO7                     : bit;
signal SDC_PWR_In                   : bit;
signal SINT_TAI                     : bit;
signal DATA_SHFT                    : std_logic_vector(15 downto 0);
signal DATA_EN_HI_SHFT              : bit;
signal DATA_EN_LO_SHFT              : bit;
signal SREQ                         : bit;
signal SLOADn                       : bit;
signal TDO                          : bit;
signal RAM_RAS0n                    : bit;
signal RAM_CAS0n                    : bit;
signal RAM_DQM0H                    : bit;
signal RAM_DQM0L                    : bit;
signal IRQ_KEYBDn                   : bit;
signal DE_I                         : bit;
signal DE_MSYNC                     : bit;
signal MULTISYNC_I                  : bit_vector(1 downto 0);
signal VIDEO_HIMODE_I               : bit;
signal DS1392_IN                    : bit;
signal DS1392_OUT                   : bit;
signal DS1392_OUT_EN                : bit;
signal PLL_LOCKS                    : bit;
signal SDATA_L_DMA                  : bit_vector(7 downto 0);
signal SDATA_R_DMA                  : bit_vector(7 downto 0);
signal SDATA_YM                     : bit_vector(7 downto 0);
signal SDATA_L                      : bit_vector(7 downto 0);
signal SDATA_R                      : bit_vector(7 downto 0);
signal ROMSEL_FC_E0n                : bit;
signal FLASH_RESET_In               : bit;
signal FLASH_WAITSTATEn             : bit;
signal BOOT_LED                     : bit;
signal LED1_I                       : bit;
signal LED2_I                       : bit;
signal PLL_ARESET                   : std_logic;
signal YM_OUT_A                     : bit;
signal YM_OUT_B                     : bit;
signal YM_OUT_C                     : bit;
signal ROM_CEn                      : bit;
signal FDRQ                         : bit;
signal FDINT                        : bit;
signal HDRQ                         : bit;
signal HDACK                        : bit;
signal HDINTn                       : bit;
signal FDCSn                        : bit;
signal HDCSn                        : bit;
signal FDD_RDn                      : bit;
signal FDD_TR00n                    : bit;
signal FDD_IPn                      : bit;
signal FDD_WPn                      : bit;
signal FDD_FDTYPE                   : bit;
signal FDD_MO                       : bit;
signal FDD_WG                       : bit;
signal FDD_WD                       : bit;
signal FDD_STEP                     : bit;
signal FDD_DIRC                     : bit;
signal SDC_EN                       : bit;
signal SDC_CSn                      : bit;
signal SDC_CLK                      : bit;
signal SDC_MISO                     : bit;
signal SDC_MOSI                     : bit;
signal SPI_CLK                      : bit;
signal SPI_MISO                     : bit;
signal SPI_MOSI                     : bit;
signal T_1S                         : bit;
signal XFF827D                      : bit_vector(7 downto 0);
signal PLL_FAULT                    : bit;
signal KEYB_TxD_I                   : bit;
begin
    -- Clock system:
    CLK_PLL_IN0 <= CLK_PLL1;
    CLK_PLL2_IN0 <= CLK_PLL2;
    CLK_CPU <= To_Bit(CLK_PLL_CPU);
    CLK_4 <= To_Bit(CLK_PLL_4x);
    CLK_2 <= To_Bit(CLK_PLL_2x);
    CLK_1 <= To_Bit(CLK_PLL_1x);
        
    RAM_CLK <= CLK_PLL_2x; -- Use the MCU clock.
    RAM_CKE <= RESET_COREn;

    KEYB_TxD <= 'Z' when RESET_CORE_Sn = '0' else
                'Z' when CONFIG(1) = '1' else
                '1' when KEYB_TxD_I = '1' else '0';

    KEY_SCAN: process
    -- Sample the RESETn and the RESET_COREn buttons
    -- about every 5ms. This provides stability against
    -- push button jitter.
    variable SCAN_TIMER    : std_logic_vector(16 downto 0);
    begin
        wait until CLK_1 = '1' and CLK_1' event;
        if SCAN_TIMER <= '1' & x"3880" then
            SCAN_TIMER := SCAN_TIMER + '1';
        else
            SCAN_TIMER := (others => '0');
            RESET_Sn <= To_Bit(RESETn);
            RESET_CORE_Sn <= RESET_COREn;
        end if;
    end process KEY_SCAN;

    PLL_RESET: process
    -- This process is responsible for resetting the PLLs
    -- during system startp.
    variable LOCK : boolean;
    begin
        wait until CLK_PLL1 = '1' and CLK_PLL1' event;
        if RESET_COREn = '0' and LOCK = false then
            LOCK := true;
            PLL_ARESET <= '1';
        elsif RESET_COREn = '1' then
            LOCK := false;
            PLL_ARESET <= '0';
        else
            PLL_ARESET <= '0';
        end if;
    end process PLL_RESET;

    PLL_LOCK_FLT: process
    -- This process provides a filter for the PLL status
    -- information.
    variable TMP : integer range 0 to 31;
    begin
        wait until CLK_PLL1 = '1' and CLK_PLL1' event;
        if (PLL1_LOCKED = '0' or PLL2_LOCKED = '0') and TMP = 0 then
            PLL_LOCKS <= '0';
            PLL_FAULT <= '1';
        elsif PLL1_LOCKED = '0' or PLL2_LOCKED = '0' then
            TMP := TMP -1;
            PLL_LOCKS <= '1';
            PLL_FAULT <= '0';
        else
            TMP := 31;
            PLL_LOCKS <= '1';
            PLL_FAULT <= '0';
        end if;
    end process PLL_LOCK_FLT;    

    -- The RESETs are as follows:
    -- RESET_Sn is the user's reset button.
    -- The RESET_CORE_Sn is the system's reset button.
    -- The RESET_BOOTn is the bootloader's reset during flash load operation.
    -- The RESET_MCUn is the memory controller's reset during RAM initialisation.
    -- PLL_LOCKS reset the system when the PLLs do not lock.
    -- RESET_EN_68K00 is the CPU reset output.
    RESET_INn <= RESET_Sn and RESET_BOOTn and RESET_MCUn and PLL_LOCKS;
    RESET_CPUn <= RESET_CORE_Sn and RESET_BOOTn and RESET_MCUn and PLL_LOCKS;
    RESETn <= '0' when RESET_EN_68K00 = '1' or FLASH_RESET_In = '0' or RESET_MCUn = '0' or PLL_LOCKS = '0' else 'Z';
    HALT_INn <= '0' when RESET_INn = '0' and RESET_EN_68K00 = '0' else '1';
    
    -------------------- Hardware specific components --------------------
    ----                                                              ----
    ---- The following components instantiate the clock phase locked  ----
    ---- loops and the video ram. These are cyclone specific and thus ----
    ---- an object of change, if other devices are used for this core.----
    ----                                                              ----
    I_SYSCLOCKS: cyclone3_pll_1
        port map(
            areset                => PLL_ARESET,
            inclk0                => CLK_PLL_IN0, -- 16MHz.
            c0                    => CLK_PLL_1x, -- 16MHz.
            c1                    => CLK_PLL_2x, -- 32MHz.
            c2                    => CLK_PLL_CPU, -- 12MHz.
            c3                    => CLK_PLL_4x, -- 64MHz.
            locked                => PLL1_LOCKED
        );

    I_AUXCLOCKS: cyclone3_pll_2
        port map(
            areset                => PLL_ARESET,
            inclk0                => CLK_PLL2_IN0, -- 16MHz.
            c0                    => CLK_PLL_256, -- 25.6MHz.
            c1                    => CLK_PLL_394, -- 39.4MHz.
            c2                    => CLK_PLL_16000, -- 16.0MHz.
            locked                => PLL2_LOCKED
        );

    ----                                                              ----
    ------------------ End hardware specific components ------------------

    P_AUX_CLOCKS: process
    -- The sound wave clock CLK_2M0 and the UART receiver,
    -- UART transmitter and sound clock CLK_0M5 are slow
    -- and therefore not possible to be provided by a PLL.
    -- Therefore this clock divider is adjusted to produce
    -- the required frequencies of 2MHz for the CLK_2M0 and
    -- 500kHz for the CLK_0M5. The Clocks are not used as
    -- clocks for d type flip-flops and therefore allowed
    -- as gated clocks.
    variable TMP: std_logic_vector(3 downto 0);
    begin
        wait until CLK_PLL_16000 = '1' and CLK_PLL_16000' event; -- 16MHz.
        TMP := TMP + '1';
        case TMP is
            when "0000" => 
                CLK_0M5 <= not CLK_0M5;
                CLK_2M0 <= not CLK_2M0;
            when "0100" | "1000" | "1100" =>
                CLK_2M0 <= not CLK_2M0;
            when others => null;
        end case;
    end process P_AUX_CLOCKS;

    P_2M4576: process
    -- This process provides the 2.4576MHz clock for the MFP
    -- timer. It is derived from a 39.4MHz PLL clock divided
    -- by 16 which results in a 2.4600 MHz clock.
    variable TMP_2M54: std_logic_vector(3 downto 0);
    begin
        wait until CLK_PLL_394 = '1' and CLK_PLL_394' event;
        TMP_2M54 := TMP_2M54 + '1';
        CLK_24576 <= To_Bit(TMP_2M54(3));
    end process P_2M4576;

    P_6M4: process
    -- This process provides the 6.4000MHz clock for the DMA
    -- sound module. It is derived from a 25.6MHz PLL clock
    -- divided by 4 which results in a 6.4000 MHz clock.
    variable TMP_6M4: std_logic_vector(1 downto 0);
    begin
        wait until CLK_PLL_256 = '1' and CLK_PLL_256' event;
        TMP_6M4 := TMP_6M4 + '1';
        SCLK_6M4 <= To_Bit(TMP_6M4(1));
    end process P_6M4;

    -- RAM bank wiring to one physical chip.
    RAM_RASn <= RAM_RAS0n;
    RAM_CASn <= RAM_CAS0n;
    RAM_DQMH <= RAM_DQM0H;
    RAM_DQML <= RAM_DQM0L;
    
    MEM_DATA_BUFFER: process(RESET_INn, CLK_2)
    -- This process is the synchronous pendant of the
    -- memory to data bus bridge buffer of the original ST
    -- machine. To work properly, the buffer is driven by
    -- a fast clock.
    begin
        if RESET_INn = '0' then
            MDAT_BUFFER <= (others => '0');
        elsif CLK_2 = '1' and CLK_2' event then
            if LATCHn = '1' then
                MDAT_BUFFER <= RAM_DATA;
            end if;
        end if;
    end process MEM_DATA_BUFFER;

    DATA <= DATA_OUT_BOOT when DATA_EN_BOOT = '1' else
            (others => '1') when RESET_BOOTn = '0' else -- Remark1.
            DATA_OUT_68K00 when DATA_EN_68K00 = '1' else
            DATA_OUT_BLT when DATA_EN_BLT = '1' else
            DATA_OUT_GLUE when DATA_EN_GLUE = '1' else
            x"FF" & DATA_OUT_MCU when DATA_EN_MCU = '1' else -- x"FF" due to pull up resistors in original hardware.
            DATA_OUT_DMA when DATA_EN_DMA = '1' else
            x"FF" & DATA_OUT_MFP when DATA_EN_MFP = '1' else
            DATA_OUT_SOUND & x"FF" when DATA_EN_SOUND = '1' else
            DATA_OUT_ACIA_I & x"FF" when DATA_EN_ACIA_I = '1' else
            x"FFF" & DATA_OUT_RP5C15 when DATA_EN_RP5C15 = '1' else
            RAM_DATA when LATCHn = '1' and RDATn = '0' else -- Remark2.
            MDAT_BUFFER when RDATn = '0' else (others => 'Z');

            -- Remark1: This prioritized logic prevents the system from
            --          any misbehaviour during reading or writing to the
            --          flash memory via the bootloader mechanism.
            -- Remark2: The RAM data is switched directly to the data bus,
            --          when the LATCH is transparent. The reason is the
            --          critical bus timing. Switching the data directly
            --          saves one CLK_2 clock period.

    -- The address space for the FLASH is decoded to enable operating system and ROM cartridge space.
    -- The ROM cartridge offset is address line 19 = '1' respective the upper half of the device.
    ADR <= '1' & ADR_OUT(18 downto 1) when ROM4n = '0' else -- Cartridge ROM FA0000 upto FAFFFF.
           '1' & ADR_OUT(18 downto 1) when ROM3n = '0' else -- Cartridge ROM FB0000 upto FBFFFF.
           "00" & ADR_OUT(17 downto 1) when ROM2n = '0' and ROMSEL_FC_E0n = '1' else -- 'Old TOS' ROM space. 
           '0' & ADR_OUT(18 downto 1) when ROM2n = '0' else -- For operating system address space E00000 to E3FFFF.
           ADR_OUT(19 downto 1); -- RAM access.

    ADR_IN <= To_BitVector(ADR_OUT);

    ADR_OUT <=     ADR_OUT_68K00 when ADR_EN_68K00 = '1' else
                ADR_OUT_BLT when ADR_EN_BLT = '1' else 
                x"0" & ADR_OUT_BOOT(19 downto 1) when ADR_EN_BOOT = '1' else (others => '1');

    FLASH_ADR22 <= To_Bit(ADR_OUT_BOOT(22)) when ADR_EN_BOOT = '1' else CONFIG(6);
    FLASH_ADR21 <= To_Bit(ADR_OUT_BOOT(21)) when ADR_EN_BOOT = '1' else CONFIG(5);
    FLASH_ADR20 <= To_Bit(ADR_OUT_BOOT(20)) when ADR_EN_BOOT = '1' else CONFIG(4);

    ROM_CEn <= ROM2n and ROM3n and ROM4n; -- The flash contains also ROM cartridge information.

    TIMER_1S: process
    variable CNT    : std_logic_vector(23 downto 0); 
    begin
        wait until CLK_1 = '1' and CLK_1' event;
        CNT := CNT + '1';
        T_1S <= To_Bit(CNT(23));
    end process TIMER_1S;
    
    LED1 <= T_1S when CONFIG(1) = '1' else 
            LED1_I when RESET_BOOTn = '1' else
            '1' when XFF827D = x"55" else
            '1' when ADR_OUT_BOOT(23) = '1' else '0';
            
    LED2 <= LED2_I when RESET_BOOTn = '1' else
            '1' when XFF827D = x"AA" else BOOT_LED;

    LED3 <= PLL_FAULT;
    
    -- Operating system ROM:
     -- Select TOS1.0 upto 1.04 address space for the lower two banks.
    with CONFIG(6 downto 4) select
        ROMSEL_FC_E0n <= '1' when "000" | "001", '0' when others;
    
    -- SD-type-RAM memory section: 
    MCU_ADR <= ADR_IN;

    -- The SHIFTER registers are read via the RAM_DATA bus. See respective ST schematics.
    RAM_DATA <= DATA when WDATn = '0' else 
                DATA_SHFT when DATA_EN_HI_SHFT = '1' and DATA_EN_LO_SHFT = '1' else
                DATA_SHFT(15 downto 8) & x"FF" when DATA_EN_HI_SHFT = '1' else
                x"FF" & DATA_SHFT(7 downto 0) when DATA_EN_LO_SHFT = '1' else (others => 'Z');

    -- Video configuration:
    MONOCHROME <= '1' when CONFIG(3 downto 2) = "00" else '0';
    
    MULTISYNC_I <= CONFIG(3 downto 2);

    P_DE_COUNT: process
    -- This flip flop provides a bisection of the DE frequency to
    -- meet a correct line counter value of the multifunction port
    -- timer B in case of the hi video modi with line doubling.
    variable LOCK    : boolean;
    begin
        wait until CLK_1 = '1' and CLK_1' event;
        if VIDEO_HIMODE_I = '0' then
            DE_MSYNC <= DE_I;
        elsif VIDEO_HIMODE_I = '1' and DE_I = '1' and LOCK = false then
            DE_MSYNC <= not DE_MSYNC;
            LOCK := true;
        elsif VIDEO_HIMODE_I = '1' and DE_I = '0' then
            LOCK := false;
        end if;
    end process P_DE_COUNT;

    -- Bus controls:
    UDS_INn <=     UDS_OUT_68K00n when UDS_OUT_EN_68K00 = '1' else
                UDS_OUT_BLTn when BUSCTRL_EN_BLT = '1' else
                UDS_OUT_GLUEn when BUSCTRL_EN_GLUE = '1' else '1';

    LDS_INn <=     LDS_OUT_68K00n when LDS_OUT_EN_68K00 = '1' else
                LDS_OUT_BLTn when BUSCTRL_EN_BLT = '1' else
                LDS_OUT_GLUEn when BUSCTRL_EN_GLUE = '1' else '1';

     -- The first condition of ASn is important for the GLUE's bus error
     -- logic. See process FLASH_WS.
    AS_INn <= '1' when FLASH_WAITSTATEn = '0' else
              AS_OUT_68K00n when AS_OUT_EN_68K00 = '1' else
              AS_OUT_BLTn when BUSCTRL_EN_BLT = '1' else
              AS_OUT_GLUEn when BUSCTRL_EN_GLUE = '1' else '1';

    RWn_IN <= RWn_OUT_68K00 when RW_OUT_EN_68K00 = '1' else
              RWn_OUT_BLT when BUSCTRL_EN_BLT = '1' else
              RWn_OUT_GLUE when BUSCTRL_EN_GLUE = '1' else '1';

    FC_IN <= To_BitVector(FC_OUT_68K00) when FC_OUT_EN_68K00 = '1' else
             To_BitVector(FC_OUT_BLT) when BUSCTRL_EN_BLT = '1' else
             FC_OUT_GLUE when BUSCTRL_EN_GLUE = '1' else "111";

    FLASH_WS: process (RESETn, CLK_1)
    -- This process provides a delay of seven clock cycles after the
    -- release of the RESETn. This is important for Suska-III-C because
    -- of the flash memory which is also resetted by the RESETn and is
    -- ready to be read at a minimum of 200ns after the release of RESETn.
    -- Without this logic, the CPU reads too fast from the flash memory
    -- when it releases a RESET_MCUn by itself.
    variable TMP: std_logic_vector(2 downto 0);
    begin
        if RESETn = '0' then
            TMP := "000";
            FLASH_WAITSTATEn <= '0';
        elsif CLK_1 = '1' and CLK_1' event then
            if TMP < "111" then
                TMP := TMP + '1';
                FLASH_WAITSTATEn <= '0';
            else
                FLASH_WAITSTATEn <= '1';
            end if;
        end if;
    end process FLASH_WS;

    SLOW_CPU: process(CLK_1, DTACK_In, ROMSEL_FC_E0n, ROM2n)
    -- For software compatibility, it is sometimes necessary to
    -- slow down the CPU. This is done by a delay of the DTACK_INn
    -- signal for operating system access. Be aware, that the DTACKn
    -- signal of the MCU may not be affected due to strong timing
    -- constraints. This feature helps to fix issues with NOP delays.
    -- The delay of the DTACK_INn causes the CPU to insert waitstates.
    variable TMP : std_logic_vector(2 downto 0);
    begin
        if CLK_1 = '1' and CLK_1' event then
            if DTACK_In = '1' then
                TMP := "000";
            elsif TMP /= "110" then
                TMP := TMP + '1';
            end if;
        end if;
        --
        case ROMSEL_FC_E0n is
            when '0' => DTACK_INn <= DTACK_In; -- Not delayed.
            when others => 
                if ROM2n = '0' and TMP = "110" then -- Slow down flash memory access.
                    DTACK_INn <= '0';
                elsif ROM2n = '0' then
                    DTACK_INn <= '1';
                else
                    DTACK_INn <= DTACK_In;
                end if;
        end case;
    end process SLOW_CPU;

    DTACK_In <= '1' when FLASH_WAITSTATEn = '0' else -- After a system reset, see process FLASH_WS.
                '0' when DTACK_OUT_BLTn = '0' or DTACK_OUT_GLUEn = '0' else
                '0' when DTACK_OUT_MCUn = '0' or DTACK_OUT_MFPn = '0' else '1';
    
    -- Bus arbitration request:
    BR_In <= BR_BLTn and BR_GLUEn;

    -- ACSI section:
    CD_IN <= CD_OUT_DMA when CD_EN_DMA = '1' else
             CD_OUT_FDC when CD_EN_FDC = '1' else
             CD_OUT_DRIVES when CD_EN_DRIVES = '1' else (others => '0');

    -- Video section:
    HSYNC_In <= not To_Bit(HSYNC);
    HSYNC <= '0' when HSYNC_On = '1' and SYNC_EN = '1' else
             '1' when HSYNC_On = '0' and SYNC_EN = '1' else 'Z';

    VSYNC_In <= not To_Bit(VSYNC);
    VSYNC <= '0' when VSYNC_On = '1' and SYNC_EN = '1' else
             '1' when VSYNC_On = '0' and SYNC_EN = '1' else 'Z';

    -- DS1392 RTC interface:
    DS1392_IN <= To_Bit(DS1392_D);
    DS1392_D <= '1' when DS1392_OUT = '1' and DS1392_OUT_EN = '1' else
                '0' when DS1392_OUT = '0' and DS1392_OUT_EN = '1' else 'Z';

    -- SD card bus:
    SDC_BSYn <= not SDC_EN;
    SDC_D(3) <= 'Z' when CONFIG(1) = '1' else
                '1' when SDC_CSn = '1' and SDC_EN = '1' else
                '0' when SDC_CSn = '0' and SDC_EN = '1' else 'Z';
    SDC_D(2 downto 1) <= "ZZ";

    SDC_MISO <= To_Bit(MISO);
    MOSI <= 'Z' when CONFIG(1) = '1' else
            '1' when SDC_MOSI = '1' and SDC_EN = '1' else
            '1' when SDC_MOSI = '0' and SDC_EN = '1' else 'Z';

    SPI_MOSI <= To_Bit(MOSI);
    MISO <= 'Z' when CONFIG(1) = '1' else
            '1' when SPI_MISO = '1' and RESET_BOOTn = '0' else
            '0' when SPI_MISO = '0' and RESET_BOOTn = '0' else 'Z';

    SPI_CLK <= To_bit(SDC_SPI_CLK);
    SDC_SPI_CLK <= 'Z' when CONFIG(1) = '1' else
                   '1' when SDC_CLK = '1' and SDC_EN = '1' else
                   '0' when SDC_CLK = '0' and SDC_EN = '1' else 'Z';

    SDC_PWRn <= '0' when CONFIG(1) = '1' else SDC_PWR_In;

    -- Sound Multiplexer:
    SDATA_L <= SDATA_L_DMA when XFF827D(7) = '1' else SDATA_YM;
    SDATA_R <= SDATA_R_DMA when XFF827D(7) = '1' else SDATA_YM;

    I_CPU: WF68K00IP_TOP_SOC
        port map(
            CLK                     => CLK_CPU,
            RESET_COREn             => RESET_CPUn,

            -- Address and Data:
            ADR_OUT                 => ADR_OUT_68K00,
            ADR_EN                  => ADR_EN_68K00,
            DATA_IN                 => DATA,
            DATA_OUT                => DATA_OUT_68K00,
            DATA_EN                 => DATA_EN_68K00,

            -- System control:
            BERRn                   => BERRn,
            RESET_INn               => RESET_INn,
            RESET_OUT_EN            => RESET_EN_68K00,
            HALT_INn                => HALT_INn,
            -- HALT_OUT_EN          =>,

            -- Processor status:
            FC_OUT                  => FC_OUT_68K00,
            FC_OUT_EN               => FC_OUT_EN_68K00,

            -- Interrupt control:
            AVECn                   => AVECn,
            IPLn                    => IPLn,

            -- Aynchronous bus control:
            DTACKn                  => DTACK_INn,
            AS_OUTn                 => AS_OUT_68K00n,
            AS_OUT_EN               => AS_OUT_EN_68K00,
            RWn_OUT                 => RWn_OUT_68K00,
            RW_OUT_EN               => RW_OUT_EN_68K00,
            UDS_OUTn                => UDS_OUT_68K00n,
            UDS_OUT_EN              => UDS_OUT_EN_68K00,
            LDS_OUTn                => LDS_OUT_68K00n,
            LDS_OUT_EN              => LDS_OUT_EN_68K00,

            -- Synchronous peripheral control:
            E                       => E_I,
            VMA_OUTn                => VMAn,
            -- VMA_OUT_EN           =>,
            VPAn                    => VPAn,

            -- Bus arbitration control:
            BRn                     => BR_In,
            BGn                     => BG_68K00n,
            BGACKn                  => BGACK_BLTn
            );

    I_BLITTER: WF101643IP_TOP_SOC
        port map(
            -- System controls:
            CLK                     => CLK_1,
            RESETn                  => RESET_INn,

            AS_INn                  => AS_INn,
            AS_OUTn                 => AS_OUT_BLTn,
            LDS_INn                 => LDS_INn,
            LDS_OUTn                => LDS_OUT_BLTn,
            UDS_INn                 => UDS_INn,
            UDS_OUTn                => UDS_OUT_BLTn,
            RWn_IN                  => RWn_IN,
            RWn_OUT                 => RWn_OUT_BLT,
            DTACK_INn               => DTACK_INn,
            DTACK_OUTn              => DTACK_OUT_BLTn,
            BERRn                   => BERRn,
            FC_IN                   => FC_IN,
            FC_OUT                  => FC_OUT_BLT,
            BUSCTRL_EN              => BUSCTRL_EN_BLT,
            INTn                    => INT_BLTn,

            -- The bus:
            ADR_IN                  => ADR_IN,
            ADR_OUT                 => ADR_OUT_BLT,
            ADR_EN                  => ADR_EN_BLT,
            DATA_IN                 => DATA,
            DATA_OUT                => DATA_OUT_BLT,
            DATA_EN                 => DATA_EN_BLT,

            -- Bus arbitration:
            BGIn                    => BG_68K00n,
            BGKIn                   => BGACK_GLUEn,
            BRn                     => BR_BLTn,
            BGACK_INn               => '1',
            BGACK_OUTn              => BGACK_BLTn,
            BGOn                    => BG_BLTn
        );

    I_GLUE: WF25915IP_TOP_V1_SOC
        port map(
            -- Clock system:
            GL_CLK                  => CLK_1,
            GL_CLK_016              => CLK_0M5,

            -- Adress decoder:
            GL_ROMSEL_FC_E0n        => ROMSEL_FC_E0n,
            EN_RAM_14MB             => '1',
            -- GL_ROM_6n            =>,
            -- GL_ROM_5n            =>, 
            GL_ROM_4n               => ROM3n,
            GL_ROM_3n               => ROM4n,
            GL_ROM_2n               => ROM2n,
            -- GL_ROM_1n            =>,
            -- GL_ROM_0n            =>,

            GL_ACIACS               => ACIA_CS_I,
            GL_MFPCSn               => MFP_CS_In,
            -- GL_SNDCSn            =>, -- Not used.
            GL_FCSn                 => FCS_In,

            GL_STE_SNDCS            => SNDCS_I,
            GL_STE_SNDIR            => SNDIR_I,

            -- RP5C15 real time clock:
            GL_STE_RTCCSn           => RP5C15_CSn,
            GL_STE_RTC_WRn          => RP5C15_WRn,
            GL_STE_RTC_RDn          => RP5C15_RDn,

            -- 6800 peripheral control:
            GL_VPAn                 => VPAn,
            GL_VMAn                 => VMAn,

            GL_DEVn                 => DEV_In,
            GL_RAMn                 => RAM_In,
            GL_DMAn                 => DMA_In,

            -- Interrupt system:
            GL_AVECn                => AVECn,
            GL_STE_FDINT            => FDINT,
            GL_STE_HDINTn           => HDINTn,
            GL_MFPINTn              => MFPINTn,
            GL_STE_EINT3n           => '1',
            GL_STE_EINT5n           => '1',
            GL_STE_EINT7n           => '1',
            GL_STE_DINTn            => DINTn,
            GL_IACKn                => IACKn,
            GL_STE_IPL2n            => IPLn(2),
            GL_STE_IPL1n            => IPLn(1),
            GL_STE_IPL0n            => IPLn(0),

            -- Video timing:
            GL_BLANKn               => BLANKn,
            GL_DE                   => DE_I,
            GL_MULTISYNC            => MULTISYNC_I,
            GL_VIDEO_HIMODE         => VIDEO_HIMODE_I,
            GL_HSYNC_INn            => HSYNC_In,
            GL_HSYNC_OUTn           => HSYNC_On,
            GL_VSYNC_INn            => VSYNC_In,
            GL_VSYNC_OUTn           => VSYNC_On,
            GL_SYNC_OUT_EN          => SYNC_EN,

            -- Bus arbitration control:
            GL_RDY_INn              => RDY_DMAn,
            GL_RDY_OUTn             => RDY_GLUEn,
            GL_BRn                  => BR_GLUEn,
            GL_BGIn                 => BG_BLTn,
            -- GL_BGOn              =>,
            GL_BGACK_INn            => BGACK_BLTn,
            GL_BGACK_OUTn           => BGACK_GLUEn,

            -- Adress and data bus:
            GL_ADDRESS              => ADR_IN,
            GL_DATA_IN              => DATA(15 downto 8),
            GL_DATA_OUT             => DATA_OUT_GLUE,
            GL_DATA_EN              => DATA_EN_GLUE,

            -- Asynchronous bus control:
            GL_RWn_IN               => RWn_IN,
            GL_RWn_OUT              => RWn_OUT_GLUE,
            GL_AS_INn               => AS_INn,
            GL_AS_OUTn              => AS_OUT_GLUEn,
            GL_UDS_INn              => UDS_INn,
            GL_UDS_OUTn             => UDS_OUT_GLUEn,
            GL_LDS_INn              => LDS_INn,
            GL_LDS_OUTn             => LDS_OUT_GLUEn,
            GL_DTACK_INn            => DTACK_INn,
            GL_DTACK_OUTn           => DTACK_OUT_GLUEn,
            GL_CTRL_EN              => BUSCTRL_EN_GLUE,

            -- System control:
            GL_RESETn               => RESET_INn,
            GL_BERRn                => BERRn,

            -- Processor function codes:
            GL_FC_IN                => FC_IN,
            GL_FC_OUT               => FC_OUT_GLUE,

            -- STE enhancements:
            -- GL_STE_FDDS          =>,
            -- GL_STE_FCCLK         =>,
            -- GL_STE_JOY_RHn       =>,
            -- GL_STE_JOY_RLn       =>,
            -- GL_STE_JOY_WL        =>,
            -- GL_STE_JOY_WEn       =>,
            -- GL_STE_BUTTONn       =>,
            GL_STE_PAD0Xn           => '1',
            GL_STE_PAD0Yn           => '1',
            GL_STE_PAD1Xn           => '1',
            GL_STE_PAD1Yn           => '1',
            -- GL_STE_PADRSTn       =>,
            GL_STE_PENn             => '1'
            -- GL_STE_SCCn          =>,
            -- GL_STE_CPROGn        =>
            );

    I_MCU: WF25912IP_SD_TOP_BTYPE_SOC
        port map(  
            CLK                     => CLK_2,
            SYS_RESET_INn           => RESET_CORE_Sn,
            SYS_RESET_OUTn          => RESET_MCUn,
            RESET_INn               => RESET_INn,

            EN_RAM_14MB             => '1',

            ASn                     => AS_INn,
            LDSn                    => LDS_INn,
            UDSn                    => UDS_INn,
            RWn                     => RWn_IN,

            ADR                     => MCU_ADR,

            RAMn                    => RAM_In,
            DMAn                    => DMA_In,
            DEVn                    => DEV_In,

            VSYNCn                  => VSYNC_On,
            DE                      => DE_I,
            VIDEO_HIMODE            => VIDEO_HIMODE_I,
            
            DCYCn                   => DCYCn,
            CMPCSn                  => CMPCSn,

            MONOCHROME              => MONOCHROME,
            
            SREQ                    => SREQ,
            SLOADn                  => SLOADn,
            SINT_TAI                => SINT_TAI,
            SINT_IO7                => SINT_IO7,
    
            BA                      => RAM_BA,
            MAD                     => RAM_ADR,
    
            WEn                     => RAM_WEn,
    
            DQM0H                   => RAM_DQM0H,
            DQM0L                   => RAM_DQM0L,
                
            RAS0n                   => RAM_RAS0n,
    
            CAS0n                   => RAM_CAS0n,
    
            RDATn                   => RDATn,
            WDATn                   => WDATn,
            LATCHn                  => LATCHn,

            DTACKn                  => DTACK_OUT_MCUn,

            DATA_IN                 => DATA(7 downto 0),
            DATA_OUT                => DATA_OUT_MCU,
            DATA_EN                 => DATA_EN_MCU
        );

    I_DMA: WF25913IP_TOP_SOC
        port map(
            -- system controls:
            RESETn                  => RESET_INn,
            CLK                     => CLK_1,

            FCSn                    => FCS_In,
            A1                      => ADR_IN(1),
            RWn                     => RWn_IN,
            RDY_INn                 => RDY_GLUEn,
            RDY_OUTn                => RDY_DMAn,
            DATA_IN                 => DATA,
            DATA_OUT                => DATA_OUT_DMA,
            DATA_EN                 => DATA_EN_DMA,

            -- ACSI mode selection:
            --DRIVE_SEL       =>,

            -- ACSI section:
            CA2                     => CA2,
            CA1                     => CA1,
            CR_Wn                   => CR_Wn,
            CD_IN                   => CD_IN,
            CD_OUT                  => CD_OUT_DMA,
            CD_EN                   => CD_EN_DMA,

            FDCSn                   => FDCSn,
            -- SDCSn                =>,
            -- SCSICSn              =>,
            HDCSn                   => HDCSn,
            FDRQ                    => FDRQ,
            HDRQ                    => HDRQ,
            ACKn                    => HDACK
        );

    I_FDC: WF1772IP_TOP_SOC
        port map(
            CLK                     => CLK_1,
            RESETn                  => RESET_INn,

            CSn                     => FDCSn,
            RWn                     => CR_Wn,
            A1                      => CA2,
            A0                      => CA1,
            DATA_IN                 => CD_IN,
            DATA_OUT                => CD_OUT_FDC,
            DATA_EN                 => CD_EN_FDC,

            RDn                     => FDD_RDn,
            TR00n                   => FDD_TR00n,
            IPn                     => FDD_IPn,
            WPRTn                   => FDD_WPn,
            DDEn                    => '0', -- Fixed to MFM.
            HDTYPE                  => FDD_FDTYPE,
            MO                      => FDD_MO,
            WG                      => FDD_WG,
            WD                      => FDD_WD,
            STEP                    => FDD_STEP,
            DIRC                    => FDD_DIRC,
            DRQ                     => FDRQ,
            INTRQ                   => FDINT
        );

    I_DRIVES: WF_SD_DRIVES
        port map(
            RESETn                  => RESET_INn,
            CLK                     => CLK_4,

            CD_IN                   => CD_IN,
            CD_OUT                  => CD_OUT_DRIVES,
            CD_EN                   => CD_EN_DRIVES,
            CR_Wn                   => CR_Wn,
            CA1                     => CA1,
    
            HDCSn                   => HDCSn,
            HDRQ                    => HDRQ,
            ACKn                    => HDACK,
            HDINTn                  => HDINTn,
    
            FDD_RDn                 => FDD_RDn,
            FDD_TR00n               => FDD_TR00n,
            FDD_IPn                 => FDD_IPn,
            FDD_WPRTn               => FDD_WPn,
            FDD_FDTYPE              => FDD_FDTYPE,
            FDD_MO                  => FDD_MO,
            FDD_WG                  => FDD_WG,
            FDD_WD                  => FDD_WD,
            FDD_STEP                => FDD_STEP,
            FDD_DIRC                => FDD_DIRC,
                
            JOY_DIS                 => JOY_DIS,
            LED_1                   => LED1_I,
            LED_2                   => LED2_I,
    
            SDC_EN                  => SDC_EN,
            SDC_MISO                => SDC_MISO,
            SDC_MOSI                => SDC_MOSI,
            SDC_CSn                 => SDC_CSn,
            SDC_SCK                 => SDC_CLK,
            SDC_CDn                 => SDC_CDn,
            SDC_WP                  => SDC_WP,
            SDC_PWRn                => SDC_PWR_In
        );

    I_SHIFTER: WF25914IP_TOP_SOC
        port map(
            CLK                     => CLK_2,
            RESETn                  => RESET_INn,

            SH_A                    => ADR_IN(6 downto 1),
            SH_D_IN                 => RAM_DATA,
            SH_D_OUT                => DATA_SHFT,
            SH_DATA_HI_EN           => DATA_EN_HI_SHFT,
            SH_DATA_LO_EN           => DATA_EN_LO_SHFT,
            SH_RWn                  => RWn_IN,
            SH_CSn                  => CMPCSn,

            MULTISYNC               => MULTISYNC_I,
            SH_LOADn                => DCYCn,
            SH_DE                   => DE_I,
            SH_BLANKn               => BLANKn,
            -- CR_1512              =>, -- Not used.
            SH_R                    => CRT_R,
            SH_G                    => CRT_G,
            SH_B                    => CRT_B,
            -- SH_MONO              =>,
            -- SH_CSYNCn            =>, -- Not used.

            SH_SCLK                 => SCLK_6M4,
            SH_FCLK                 => FCLK,
            SH_SLOADn               => SLOADn,
            SH_SREQ                 => SREQ,
            SH_SDATA_L              => SDATA_L_DMA,
            SH_SDATA_R              => SDATA_R_DMA,

            -- SH_MWK               =>,
            -- SH_MWD               =>,
            -- SH_MWEn              =>,

            xFF827E_D               => XFF827D
        );

    I_MFP: WF68901IP_TOP_SOC
        port map(  
            -- System control:
            CLK                     => CLK_1,
            RESETn                  => RESET_INn,

            -- Asynchronous bus control:
            DSn                     => LDS_INn,
            CSn                     => MFP_CS_In,
            RWn                     => RWn_IN,
            DTACKn                  => DTACK_OUT_MFPn,
    
            -- Data and Adresses:
            RS                      => ADR_IN(5 downto 1),
            DATA_IN                 => DATA(7 downto 0),
            DATA_OUT                => DATA_OUT_MFP,
            DATA_EN                 => DATA_EN_MFP,
            GPIP_IN(7)              => SINT_IO7,
            GPIP_IN(6)              => '0',
            GPIP_IN(5)              => DINTn,
            GPIP_IN(4)              => IRQ_KEYBDn,
            GPIP_IN(3)              => INT_BLTn,
            GPIP_IN(2)              => '0',
            GPIP_IN(1)              => '0',
            GPIP_IN(0)              => '0',
            -- GPIP_OUT             =>, -- Not used; all GPIPs are direction input.
            -- GPIP_EN              =>, -- Not used; all GPIPs are direction input.
    
            -- Interrupt control:
            IACKn                   => IACKn,
            IEIn                    => '0',
            -- IEOn                 =>, -- Not used.
            IRQn                    => MFPINTn,
    
            -- Timers and timer control:
            XTAL1                   => CLK_24576,
            TAI                     => SINT_TAI,
            TBI                     => DE_MSYNC,
            -- TAO                  =>, -- Not used.
            -- TBO                  =>, -- Not used.
            -- TCO                  =>, -- Not used.
            TDO                     => TDO,
    
            -- Serial I/O control:
            RC                      => TDO,
            TC                      => TDO,
            SI                      => '1'
            -- SO                   =>,
            -- SO_EN                =>

            -- DMA control:
            -- RRn                  =>, -- Not used.
            -- TRn                  => -- Not used.
        );

    I_SOUND: WF2149IP_DIGOUT_TOP_SOC
        port map(
            SYS_CLK                 => CLK_1,
            RESETn                  => RESET_INn,

            WAV_CLK                 => CLK_2M0,
            SELn                    => '1',

            BDIR                    => SNDIR_I,
            BC2                     => '1',
            BC1                     => SNDCS_I,

            A9n                     => '0',
            A8                      => '1',
            DA_IN                   => DATA(15 downto 8),
            DA_OUT                  => DATA_OUT_SOUND,
            DA_EN                   => DATA_EN_SOUND,

            IO_A_IN                 => x"00", -- All port pins are dedicated outputs.
            -- IO_A_OUT(7)          =>, -- Not used so far.
            -- IO_A_OUT(6)          =>,
            -- IO_A_OUT(5)          =>,
            -- IO_A_OUT(4)          =>,
            -- IO_A_OUT(3)          =>,
            -- IO_A_OUT(2)          =>,
            -- IO_A_OUT(1)          =>,
            -- IO_A_OUT(0)          =>,
            -- IO_A_EN              =>, -- Not required.
            IO_B_IN                 => x"00",
            -- IO_B_OUT             =>,
            -- IO_B_EN              =>,

            OUT_ALL                 => SDATA_YM
        );

    I_ACIA_KEYBOARD: WF6850IP_TOP_SOC
      port map(
            CLK                     => CLK_1,
            RESETn                  => RESET_INn,

            CS2n                    => ADR_IN(2),
            CS1                     => '1',
            CS0                     => ACIA_CS_I,
            E                       => E_I,
            RWn                     => RWn_IN,
            RS                      => ADR_IN(1),

            DATA_IN                 => DATA(15 downto 8),
            DATA_OUT                => DATA_OUT_ACIA_I,
            DATA_EN                 => DATA_EN_ACIA_I,

            TXCLK                   => CLK_0M5,
            RXCLK                   => CLK_0M5,
            RXDATA                  => KEYB_RxD,

            CTSn                    => '0', -- In original ST machines wired to GND.
            DCDn                    => '0', -- In original ST machines wired to GND.
        
            IRQn                    => IRQ_KEYBDn,
            TXDATA                  => KEYB_TxD_I
            --RTSn                  => -- Not used.
        );                                              

    I_RTC5C15: WF5C15_139xIP_TOP
        port map(
            CLK                     => CLK_1,
            RESETn                  => RESET_INn,

            -- The bus interface:
            ADR                     => ADR_IN(4 downto 1),
            DATA_IN                 => DATA(3 downto 0),
            DATA_OUT                => DATA_OUT_RP5C15,
            DATA_EN                 => DATA_EN_RP5C15,
            CS                      => '1',
            CSn                     => RP5C15_CSn,
            WRn                     => RP5C15_WRn,
            RDn                     => RP5C15_RDn,

            -- The SPI lines:
            SPI_IN                  => DS1392_IN,
            SPI_OUT                 => DS1392_OUT,
            SPI_EN                  => DS1392_OUT_EN,
            SPI_SCL                 => DS1392_SCL,
            SPI_CE                  => DS1392_CE
        );

    I_AUDIODAC: WF_AUDIO_DAC
        port map(
            CLK                     => CLK_1, -- 16MHz.
            RESETn                  => RESET_INn,

            FCLK                    => FCLK,

            SDATA_L                 => SDATA_L,
            SDATA_R                 => SDATA_R,
            DAC_SCLK                => DAC_SCLK,
            DAC_SDATA               => DAC_SDATA,
            DAC_SYNCn               => DAC_SYNCn,
            DAC_LDACn               => DAC_LDACn
        );

    I_FLASHBOOT: WF_FLASHBOOT
        port map(
            CLK                     => CLK_1, -- 16MHz.
            PLL_LOCK                => PLL_LOCKS,
            RESET_COREn             => RESET_CORE_Sn,
            RESET_INn               => RESET_Sn,
            RESET_OUTn              => RESET_BOOTn,

            ROM_CEn                 => ROM_CEn,
            ADR_OUT                 => ADR_OUT_BOOT, -- ADR(23) currently not in use (LED1).
            ADR_EN                  => ADR_EN_BOOT,
            DATA_IN                 => DATA,
            DATA_OUT                => DATA_OUT_BOOT,
            DATA_EN                 => DATA_EN_BOOT,
            FLASH_RDY               => FLASH_RDY,
            FLASH_RESETn            => FLASH_RESET_In,
            FLASH_WEn               => FLASH_WEn,
            FLASH_OEn               => FLASH_OEn,
            FLASH_CEn               => FLASH_CEn,
            SPI_CLK                 => SPI_CLK,
            SPI_DIN                 => SPI_MOSI,
            SPI_DOUT                => SPI_MISO,
            BOOT_ACK                => BOOT_ACK,
            BOOT_REQ                => BOOT_REQ,
            BOOT_LED                => BOOT_LED
        );
end architecture STRUCTURE;
-- To Do:
--  Pin Assignments: MOSI / MISO tauschen.
--  Prototypen: and SD-Karte MOSI/MISO umverdrahten.